from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, ClassVar, Optional, cast

import pytest

from chik.consensus.block_record import BlockRecord
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.full_block import FullBlock
from chik.util.augmented_chain import AugmentedBlockchain
from chik.util.errors import Err
from chik.util.ints import uint32


@dataclass
class NullBlockchain:
    if TYPE_CHECKING:
        from chik.consensus.blockchain_interface import BlocksProtocol

        _protocol_check: ClassVar[BlocksProtocol] = cast("NullBlockchain", None)

    added_blocks: set[bytes32] = field(default_factory=set)
    heights: dict[uint32, bytes32] = field(default_factory=dict)

    # BlocksProtocol
    async def lookup_block_generators(self, header_hash: bytes32, generator_refs: set[uint32]) -> dict[uint32, bytes]:
        raise ValueError(Err.GENERATOR_REF_HAS_NO_GENERATOR)  # pragma: no cover

    async def get_block_record_from_db(self, header_hash: bytes32) -> Optional[BlockRecord]:
        return None  # pragma: no cover

    def add_block_record(self, block_record: BlockRecord) -> None:
        self.added_blocks.add(block_record.header_hash)

    # BlockRecordsProtocol
    def try_block_record(self, header_hash: bytes32) -> Optional[BlockRecord]:
        return None  # pragma: no cover

    def block_record(self, header_hash: bytes32) -> BlockRecord:
        raise KeyError("no block records in NullBlockchain")  # pragma: no cover

    def height_to_block_record(self, height: uint32) -> BlockRecord:
        raise ValueError("Height is not in blockchain")

    def height_to_hash(self, height: uint32) -> Optional[bytes32]:
        return self.heights.get(height)

    def contains_block(self, header_hash: bytes32) -> bool:
        return False  # pragma: no cover

    def contains_height(self, height: uint32) -> bool:
        return height in self.heights.keys()

    async def prev_block_hash(self, header_hashes: list[bytes32]) -> list[bytes32]:
        raise KeyError("no block records in NullBlockchain")  # pragma: no cover


@dataclass
class FakeBlockRecord:
    height: uint32
    header_hash: bytes32
    prev_hash: bytes32


def BR(b: FullBlock) -> BlockRecord:
    ret = FakeBlockRecord(b.height, b.header_hash, b.prev_header_hash)
    return ret  # type: ignore[return-value]


@pytest.mark.anyio
@pytest.mark.limit_consensus_modes(reason="save time")
async def test_augmented_chain(default_10000_blocks: list[FullBlock]) -> None:
    blocks = default_10000_blocks
    # this test blockchain is expected to have block generators at these
    # heights:
    # 2, 3, 4, 5, 6, 7, 9, 10, 11, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23,
    # 24, 25, 26, 28

    null = NullBlockchain()
    abc = AugmentedBlockchain(null)

    # before adding anything to the augmented blockchain, make sure we just pass
    # through all requests
    with pytest.raises(ValueError, match="Height is not in blockchain"):
        abc.height_to_block_record(uint32(1))

    with pytest.raises(KeyError):
        abc.block_record(blocks[2].header_hash)

    with pytest.raises(KeyError):
        await abc.prev_block_hash([blocks[2].header_hash])

    with pytest.raises(ValueError, match="Err.GENERATOR_REF_HAS_NO_GENERATOR"):
        await abc.lookup_block_generators(blocks[3].header_hash, {uint32(3)})

    block_records = []

    # now add some blocks
    for b in blocks[:5]:
        block_records.append(BR(b))
        abc.add_extra_block(b, BR(b))

    assert abc.height_to_block_record(uint32(1)) == block_records[1]

    with pytest.raises(ValueError, match="Err.GENERATOR_REF_HAS_NO_GENERATOR"):
        await abc.lookup_block_generators(blocks[10].header_hash, {uint32(3), uint32(10)})

    # block 1 exists in the chain, but it doesn't have a generator
    with pytest.raises(ValueError, match="Err.GENERATOR_REF_HAS_NO_GENERATOR"):
        await abc.lookup_block_generators(blocks[1].header_hash, {uint32(1)})

    expect_gen = blocks[2].transactions_generator
    assert expect_gen is not None
    assert await abc.lookup_block_generators(blocks[5].prev_header_hash, {uint32(2)}) == {uint32(2): bytes(expect_gen)}

    for i in range(1, 5):
        assert await abc.prev_block_hash([blocks[i].header_hash]) == [blocks[i - 1].header_hash]

    for i in range(5):
        assert abc.block_record(blocks[i].header_hash) == block_records[i]
        assert abc.try_block_record(blocks[i].header_hash) == block_records[i]
        assert abc.height_to_hash(uint32(i)) == blocks[i].header_hash
        assert await abc.prev_block_hash([blocks[i].header_hash]) == [blocks[i].prev_header_hash]
        assert abc.contains_block(blocks[i].header_hash) is True
        assert await abc.get_block_record_from_db(blocks[i].header_hash) == block_records[i]
        assert abc.contains_height(uint32(i))

    for i in range(5, 10):
        assert abc.height_to_hash(uint32(i)) is None
        assert not abc.contains_block(blocks[i].header_hash)
        assert not await abc.get_block_record_from_db(blocks[i].header_hash)
        assert not abc.contains_height(uint32(i))

    assert abc.height_to_hash(uint32(5)) is None
    null.heights = {uint32(5): blocks[5].header_hash}
    assert abc.height_to_hash(uint32(5)) == blocks[5].header_hash

    # if we add blocks to cache that are already augmented into the chain, the
    # augmented blocks should be removed
    assert len(abc._extra_blocks) == 5
    for b in blocks[:5]:
        abc.add_block_record(BR(b))
    assert len(abc._extra_blocks) == 0
    assert null.added_blocks == {br.header_hash for br in blocks[:5]}



from __future__ import annotations

import json
import subprocess
import tempfile
from typing import Optional

import click
import packaging.version

from chik.daemon.server import executable_for_service
from chik.util.timing import adjusted_timeout


def check_plotter(plotter: list[str], expected_output: bytes, specify_tmp: bool = True) -> None:
    with tempfile.TemporaryDirectory() as path:
        tmp_dir = []
        if specify_tmp:
            tmp_dir = ["--tmp_dir", path]
        process = subprocess.Popen(
            [executable_for_service("chik"), "plotters", *plotter, *tmp_dir, "--final_dir", path],
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        out: Optional[bytes]
        err: Optional[bytes]
        try:
            out, err = process.communicate(timeout=2)
        except subprocess.TimeoutExpired as e:
            err = e.stderr
            out = e.stdout
        else:
            print(repr(err))
            print(repr(out))
            assert False, "expected to time out"
        finally:
            process.kill()
            process.communicate()

        assert err is None, repr(err)
        assert out is not None
        assert out.startswith(expected_output), repr(out)


@click.group("installers", help="Installer related helpers such as installer testing")
def installers_group() -> None:
    pass


@installers_group.command(name="test")
@click.option("--expected-chik-version", "expected_chik_version_str", required=True)
@click.option("--require-madmax/--require-no-madmax", "require_madmax", default=True)
def test_command(expected_chik_version_str: str, require_madmax: bool) -> None:
    print("testing installed executables")
    expected_chik_version = packaging.version.Version(expected_chik_version_str)

    args = [executable_for_service("chik"), "version"]
    print(f"launching: {args}")
    chik_version_process = subprocess.run(
        args,
        capture_output=True,
        encoding="utf-8",
        timeout=adjusted_timeout(30),
        check=False,
    )
    assert chik_version_process.returncode == 0
    assert chik_version_process.stderr == ""

    chik_version = packaging.version.Version(chik_version_process.stdout)
    print(chik_version)
    assert chik_version == expected_chik_version, f"{chik_version} != {expected_chik_version}"

    args = [executable_for_service("chik"), "plotters", "version"]
    print(f"launching: {args}")
    plotter_version_process = subprocess.run(
        args,
        capture_output=True,
        encoding="utf-8",
        timeout=adjusted_timeout(30),
        check=False,
    )

    print()
    print(plotter_version_process.stdout)
    print()
    print(plotter_version_process.stderr)
    print()

    assert plotter_version_process.returncode == 0
    assert plotter_version_process.stderr == ""

    found_start = False
    plotter_versions: dict[str, packaging.version.Version] = {}
    for line in plotter_version_process.stdout.splitlines():
        if line.startswith("chikpos:"):
            found_start = True

        if not found_start:
            continue

        plotter, version = (segment.strip() for segment in line.split(":", maxsplit=1))
        plotter_versions[plotter] = packaging.version.Version(version)

    print(json.dumps({plotter: str(version) for plotter, version in plotter_versions.items()}, indent=4))
    expected = {"chikpos", "bladebit"}

    if require_madmax:
        expected.add("madmax")

    assert plotter_versions.keys() == expected, f"{expected=}"

    # TODO: figure out a better test, these actually start plots which can use up disk
    #       space too fast

    # check_plotter(plotter=["chikpos"], expected_output=b"\nStarting plotting progress")
    # check_plotter(plotter=["madmax"], expected_output=b"Multi-threaded pipelined Chik")
    # check_plotter(plotter=["bladebit", "diskplot", "--compress", "0"], expected_output=b"\nBladebit Chik Plotter")
    # check_plotter(plotter=["bladebit", "cudaplot", "--compress", "0"], expected_output=b"\nBladebit Chik Plotter")
    # check_plotter(
    #     plotter=["bladebit", "ramplot", "--compress", "0"],
    #     expected_output=b"\nBladebit Chik Plotter",
    #     specify_tmp_dir=False,
    # )

    args = [executable_for_service("chik"), "init"]
    print(f"launching: {args}")
    subprocess.run(
        args,
        check=True,
        timeout=adjusted_timeout(30),
    )



# TODO: this should not exist, it is a bad task group that doesn't require
#       any responsibility of the requestor.

from __future__ import annotations

import asyncio
import dataclasses
import logging
import typing

T = typing.TypeVar("T")

logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class _TaskInfo:
    task: asyncio.Task[object]
    # retained for potential debugging use
    known_unreferenced: bool

    def __str__(self) -> str:
        return self.task.get_name()


@dataclasses.dataclass
class _TaskReferencer:
    """Holds strong references to tasks until they are done.  This compensates for
    asyncio holding only weak references.  This should be replaced by patterns using
    task groups such as from anyio.
    """

    tasks: dict[asyncio.Task[object], _TaskInfo] = dataclasses.field(default_factory=dict)

    def create_task(
        self,
        coroutine: typing.Coroutine[object, object, T],
        *,
        name: typing.Optional[str] = None,
        known_unreferenced: bool = False,
    ) -> asyncio.Task[T]:
        task = asyncio.create_task(coro=coroutine, name=name)  # noqa: TID251
        task.add_done_callback(self._task_done)

        self.tasks[task] = _TaskInfo(task=task, known_unreferenced=known_unreferenced)

        return task

    def _task_done(self, task: asyncio.Task[object]) -> None:
        # TODO: consider collecting results and logging errors
        try:
            del self.tasks[task]
        except KeyError:
            logger.warning("Task not found in task referencer: %s", task)


_global_task_referencer = _TaskReferencer()

create_referenced_task = _global_task_referencer.create_task



# https://github.com/bitcoin/bitcoin/blob/5b6f0f31fa6ce85db3fb7f9823b1bbb06161ae32/src/policy/fees.h
from __future__ import annotations

MIN_FEE_RATE = 0  # Value of first bucket
INITIAL_STEP = 5.0  # First bucket after zero value
MAX_FEE_RATE = 40000000.0  # Mojo per 1000 cost unit
INFINITE_FEE_RATE = 1000000000.0

STEP_SIZE = 1.05  # bucket increase by 1.05

# Track confirm delays up to SHORT_BLOCK_PERIOD blocks for short horizon
SHORT_BLOCK_PERIOD = 12  # 3
SHORT_SCALE = 1

# Track confirm delays up to MED_BLOCK_PERIOD blocks for medium horizon
MED_BLOCK_PERIOD = 24  # 15
MED_SCALE = 2

# Track confirm delays up to LONG_BLOCK_PERIOD blocks for long horizon
LONG_BLOCK_PERIOD = 42  # 15
LONG_SCALE = 24  # 4

SECONDS_PER_BLOCK = 40

SHORT_DECAY = 0.962
MED_DECAY = 0.9952
LONG_DECAY = 0.99931

HALF_SUCCESS_PCT = 0.6  # Require 60 % success rate for target confirmations
SUCCESS_PCT = 0.85  # Require 85 % success rate for target confirmations
DOUBLE_SUCCESS_PCT = 0.95  # Require 95 % success rate for target confirmations

# Require an avg of SUFFICIENT_FEE_TXS tx in the combined fee rate bucket per block to have stat. significance
SUFFICIENT_FEE_TXS = 0.01

FEE_ESTIMATOR_VERSION = 1

OLDEST_ESTIMATE_HISTORY = 6 * 1008



from __future__ import annotations

import asyncio
from typing import Optional

import pytest

from chik.util.limited_semaphore import LimitedSemaphore, LimitedSemaphoreFullError
from chik.util.task_referencer import create_referenced_task


@pytest.mark.anyio
async def test_stuff() -> None:
    active_limit = 2
    waiting_limit = 4
    total_limit = active_limit + waiting_limit
    beyond_limit = 3
    semaphore = LimitedSemaphore.create(active_limit=active_limit, waiting_limit=waiting_limit)
    finish_event = asyncio.Event()

    async def acquire(entered_event: Optional[asyncio.Event] = None) -> None:
        async with semaphore.acquire():
            assert entered_event is not None
            entered_event.set()
            await finish_event.wait()

    entered_events = [asyncio.Event() for _ in range(active_limit)]
    waiting_events = [asyncio.Event() for _ in range(waiting_limit)]
    failed_events = [asyncio.Event() for _ in range(beyond_limit)]

    entered_tasks = [create_referenced_task(acquire(entered_event=event)) for event in entered_events]
    waiting_tasks = [create_referenced_task(acquire(entered_event=event)) for event in waiting_events]

    await asyncio.gather(*(event.wait() for event in entered_events))
    assert all(event.is_set() for event in entered_events)
    assert all(not event.is_set() for event in waiting_events)

    assert semaphore._available_count == 0

    failure_tasks = [create_referenced_task(acquire()) for _ in range(beyond_limit)]

    failure_results = await asyncio.gather(*failure_tasks, return_exceptions=True)
    assert [str(error) for error in failure_results] == [str(LimitedSemaphoreFullError())] * beyond_limit
    assert all(not event.is_set() for event in failed_events)

    assert semaphore._available_count == 0

    finish_event.set()
    success_results = await asyncio.gather(*entered_tasks, *waiting_tasks)
    assert all(event.is_set() for event in waiting_events)
    assert success_results == [None] * total_limit

    assert semaphore._available_count == total_limit



from __future__ import annotations

import itertools
from dataclasses import dataclass, field
from typing import Generic, TypeVar

from chik.simulator.ssl_certs_1 import SSL_TEST_NODE_CERTS_AND_KEYS_1, SSL_TEST_PRIVATE_CA_CERT_AND_KEY_1
from chik.simulator.ssl_certs_2 import SSL_TEST_NODE_CERTS_AND_KEYS_2, SSL_TEST_PRIVATE_CA_CERT_AND_KEY_2
from chik.simulator.ssl_certs_3 import SSL_TEST_NODE_CERTS_AND_KEYS_3, SSL_TEST_PRIVATE_CA_CERT_AND_KEY_3
from chik.simulator.ssl_certs_4 import SSL_TEST_NODE_CERTS_AND_KEYS_4, SSL_TEST_PRIVATE_CA_CERT_AND_KEY_4
from chik.simulator.ssl_certs_5 import SSL_TEST_NODE_CERTS_AND_KEYS_5, SSL_TEST_PRIVATE_CA_CERT_AND_KEY_5
from chik.simulator.ssl_certs_6 import SSL_TEST_NODE_CERTS_AND_KEYS_6, SSL_TEST_PRIVATE_CA_CERT_AND_KEY_6
from chik.simulator.ssl_certs_7 import SSL_TEST_NODE_CERTS_AND_KEYS_7, SSL_TEST_PRIVATE_CA_CERT_AND_KEY_7
from chik.simulator.ssl_certs_8 import SSL_TEST_NODE_CERTS_AND_KEYS_8, SSL_TEST_PRIVATE_CA_CERT_AND_KEY_8
from chik.simulator.ssl_certs_9 import SSL_TEST_NODE_CERTS_AND_KEYS_9, SSL_TEST_PRIVATE_CA_CERT_AND_KEY_9
from chik.simulator.ssl_certs_10 import SSL_TEST_NODE_CERTS_AND_KEYS_10, SSL_TEST_PRIVATE_CA_CERT_AND_KEY_10

# ---------------------------------------------------------------------------
# NOTE:
# Use tests/util/gen_ssl_certs.py to generate additional SSL certs and keys
#
# EXAMPLE:
# $ python3 tests/util/gen_ssl_certs.py --suffix 123 > ssl_certs_123.py
# ---------------------------------------------------------------------------

_T_SSLTestCollateral = TypeVar("_T_SSLTestCollateral", bound="SSLTestCollateralTracker")


@dataclass
class SSLTestCollateralTracker:
    in_use: bool = field(default=False, init=False)

    def mark_in_use(self) -> None:
        self.in_use = True

    def mark_not_in_use(self) -> None:
        self.in_use = False


@dataclass
class SSLTestCACertAndPrivateKey(SSLTestCollateralTracker):
    cert_and_key: tuple[bytes, bytes]


@dataclass
class SSLTestNodeCertsAndKeys(SSLTestCollateralTracker):
    certs_and_keys: dict[str, dict[str, dict[str, bytes]]]


@dataclass
class SSLTestCollateralWrapper(Generic[_T_SSLTestCollateral]):
    collateral: _T_SSLTestCollateral

    def __post_init__(self) -> None:
        if self.collateral.in_use:
            print("  WARNING: Reusing SSL Test collateral that is currently in use")
        self.collateral.mark_in_use()

    def __del__(self) -> None:
        self.collateral.mark_not_in_use()


# ---------------------------------------------------------------------------
# Private CA certs/keys
# ---------------------------------------------------------------------------

SSL_TEST_PRIVATE_CA_CERTS_AND_KEYS: list[SSLTestCACertAndPrivateKey] = [
    SSLTestCACertAndPrivateKey(SSL_TEST_PRIVATE_CA_CERT_AND_KEY_1),
    SSLTestCACertAndPrivateKey(SSL_TEST_PRIVATE_CA_CERT_AND_KEY_2),
    SSLTestCACertAndPrivateKey(SSL_TEST_PRIVATE_CA_CERT_AND_KEY_3),
    SSLTestCACertAndPrivateKey(SSL_TEST_PRIVATE_CA_CERT_AND_KEY_4),
    SSLTestCACertAndPrivateKey(SSL_TEST_PRIVATE_CA_CERT_AND_KEY_5),
    SSLTestCACertAndPrivateKey(SSL_TEST_PRIVATE_CA_CERT_AND_KEY_6),
    SSLTestCACertAndPrivateKey(SSL_TEST_PRIVATE_CA_CERT_AND_KEY_7),
    SSLTestCACertAndPrivateKey(SSL_TEST_PRIVATE_CA_CERT_AND_KEY_8),
    SSLTestCACertAndPrivateKey(SSL_TEST_PRIVATE_CA_CERT_AND_KEY_9),
    SSLTestCACertAndPrivateKey(SSL_TEST_PRIVATE_CA_CERT_AND_KEY_10),
]

# ---------------------------------------------------------------------------
# Node -> cert/key mappings
# ---------------------------------------------------------------------------

SSL_TEST_NODE_CERTS_AND_KEYS: list[SSLTestNodeCertsAndKeys] = [
    SSLTestNodeCertsAndKeys(SSL_TEST_NODE_CERTS_AND_KEYS_1),
    SSLTestNodeCertsAndKeys(SSL_TEST_NODE_CERTS_AND_KEYS_2),
    SSLTestNodeCertsAndKeys(SSL_TEST_NODE_CERTS_AND_KEYS_3),
    SSLTestNodeCertsAndKeys(SSL_TEST_NODE_CERTS_AND_KEYS_4),
    SSLTestNodeCertsAndKeys(SSL_TEST_NODE_CERTS_AND_KEYS_5),
    SSLTestNodeCertsAndKeys(SSL_TEST_NODE_CERTS_AND_KEYS_6),
    SSLTestNodeCertsAndKeys(SSL_TEST_NODE_CERTS_AND_KEYS_7),
    SSLTestNodeCertsAndKeys(SSL_TEST_NODE_CERTS_AND_KEYS_8),
    SSLTestNodeCertsAndKeys(SSL_TEST_NODE_CERTS_AND_KEYS_9),
    SSLTestNodeCertsAndKeys(SSL_TEST_NODE_CERTS_AND_KEYS_10),
]


ssl_test_private_ca_certs_and_keys_gen = (
    SSL_TEST_PRIVATE_CA_CERTS_AND_KEYS[idx]
    for idx in itertools.cycle([*range(len(SSL_TEST_PRIVATE_CA_CERTS_AND_KEYS))])
)


def get_next_private_ca_cert_and_key() -> SSLTestCollateralWrapper[SSLTestCACertAndPrivateKey]:
    return SSLTestCollateralWrapper(next(ssl_test_private_ca_certs_and_keys_gen))


ssl_test_certs_and_keys_gen = (
    SSL_TEST_NODE_CERTS_AND_KEYS[idx] for idx in itertools.cycle([*range(len(SSL_TEST_NODE_CERTS_AND_KEYS))])
)


def get_next_nodes_certs_and_keys() -> SSLTestCollateralWrapper[SSLTestNodeCertsAndKeys]:
    return SSLTestCollateralWrapper(next(ssl_test_certs_and_keys_gen))



from __future__ import annotations

from dataclasses import dataclass

from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.mempool_item import BundleCoinSpend
from chik.types.spend_bundle import SpendBundle
from chik.types.spend_bundle_conditions import SpendBundleConditions
from chik.util.ints import uint32


@dataclass(frozen=True)
class InternalMempoolItem:
    spend_bundle: SpendBundle
    conds: SpendBundleConditions
    height_added_to_mempool: uint32
    # Map of coin ID to coin spend data between the bundle and its NPCResult
    bundle_coin_spends: dict[bytes32, BundleCoinSpend]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional

from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.ints import uint32
from chik.util.streamable import Streamable, recurse_jsonify, streamable


def dict_with_types(d: dict[str, Any]) -> dict[str, Any]:
    return {k: (v, type(v)) for k, v in d.items()}


def test_primitives() -> None:
    @streamable
    @dataclass(frozen=True)
    class PrimitivesTest(Streamable):
        a: uint32
        b: Optional[str]
        c: str
        d: bytes
        e: bytes32
        f: bool

    t1 = PrimitivesTest(
        uint32(123),
        None,
        "foobar",
        b"\0\1\0\1",
        bytes32(range(32)),
        True,
    )

    assert dict_with_types(t1.to_json_dict()) == {
        "a": (123, int),
        "b": (None, type(None)),
        "c": ("foobar", str),
        "d": ("0x00010001", str),
        "e": ("0x000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f", str),
        "f": (True, bool),
    }

    t2 = PrimitivesTest(
        uint32(0),
        "set optional",
        "foobar",
        b"\0\1",
        bytes32.zeros,
        False,
    )

    assert dict_with_types(t2.to_json_dict()) == {
        "a": (0, int),
        "b": ("set optional", str),
        "c": ("foobar", str),
        "d": ("0x0001", str),
        "e": ("0x0000000000000000000000000000000000000000000000000000000000000000", str),
        "f": (False, bool),
    }


def test_list() -> None:
    @streamable
    @dataclass(frozen=True)
    class ListTest(Streamable):
        d: list[str]

    t = ListTest(["foo", "bar"])

    assert t.to_json_dict() == {"d": ["foo", "bar"]}


def test_tuple() -> None:
    @streamable
    @dataclass(frozen=True)
    class TupleTest(Streamable):
        d: tuple[str, uint32, str]

    t = TupleTest(("foo", uint32(123), "bar"))

    assert t.to_json_dict() == {"d": ["foo", 123, "bar"]}


@streamable
@dataclass(frozen=True)
class NestedWithTupleInner(Streamable):
    a: tuple[str, uint32, str]
    b: bytes


@streamable
@dataclass(frozen=True)
class NestedWithTupleOuter(Streamable):
    a: tuple[NestedWithTupleInner, uint32, str]


def test_nested_with_tuple() -> None:
    t = NestedWithTupleOuter(
        (NestedWithTupleInner(("foo", uint32(123), "bar"), bytes([0x13, 0x37])), uint32(321), "baz")
    )

    assert t.to_json_dict() == {"a": [{"a": ["foo", 123, "bar"], "b": "0x1337"}, 321, "baz"]}


@streamable
@dataclass(frozen=True)
class NestedWithListInner(Streamable):
    a: uint32
    b: bytes


@streamable
@dataclass(frozen=True)
class NestedWithListOuter(Streamable):
    a: list[NestedWithListInner]


def test_nested_with_list() -> None:
    t = NestedWithListOuter([NestedWithListInner(uint32(123), bytes([0x13, 0x37]))])

    assert t.to_json_dict() == {"a": [{"a": 123, "b": "0x1337"}]}


@streamable
@dataclass(frozen=True)
class TestNestedInner(Streamable):
    a: tuple[str, uint32, str]
    b: bytes


@streamable
@dataclass(frozen=True)
class TestNestedOuter(Streamable):
    a: TestNestedInner


def test_nested() -> None:
    t = TestNestedOuter(TestNestedInner(("foo", uint32(123), "bar"), bytes([0x13, 0x37])))

    assert t.to_json_dict() == {"a": {"a": ["foo", 123, "bar"], "b": "0x1337"}}


def test_recurse_jsonify() -> None:
    d = {"a": "foo", "b": bytes([0x13, 0x37]), "c": [uint32(1), uint32(2)], "d": {"bar": None}}
    assert recurse_jsonify(d) == {"a": "foo", "b": "0x1337", "c": [1, 2], "d": {"bar": None}}



from __future__ import annotations

from pathlib import Path

from chik_rs import Coin

from chik._tests.cmds.cmd_test_utils import TestRpcClients, TestWalletRpcClient, logType, run_cli_command_and_assert
from chik._tests.cmds.wallet.test_consts import FINGERPRINT_ARG, STD_TX, STD_UTX, get_bytes32
from chik.rpc.wallet_request_types import (
    VCAddProofs,
    VCGet,
    VCGetList,
    VCGetListResponse,
    VCGetProofsForRoot,
    VCGetProofsForRootResponse,
    VCGetResponse,
    VCMint,
    VCMintResponse,
    VCProofsRPC,
    VCProofWithHash,
    VCRecordWithCoinID,
    VCRevoke,
    VCRevokeResponse,
    VCSpend,
    VCSpendResponse,
)
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.bech32m import encode_puzzle_hash
from chik.util.ints import uint32, uint64
from chik.wallet.conditions import ConditionValidTimes
from chik.wallet.lineage_proof import LineageProof
from chik.wallet.transaction_record import TransactionRecord
from chik.wallet.util.tx_config import DEFAULT_TX_CONFIG, TXConfig
from chik.wallet.vc_wallet.vc_drivers import VCLineageProof, VerifiedCredential
from chik.wallet.vc_wallet.vc_store import VCRecord

test_condition_valid_times: ConditionValidTimes = ConditionValidTimes(min_time=uint64(100), max_time=uint64(150))
# VC Commands


def test_vcs_mint(capsys: object, get_test_cli_clients: tuple[TestRpcClients, Path]) -> None:
    test_rpc_clients, root_dir = get_test_cli_clients

    # set RPC Client
    class VcsMintRpcClient(TestWalletRpcClient):
        async def vc_mint(
            self,
            request: VCMint,
            tx_config: TXConfig,
            timelock_info: ConditionValidTimes = ConditionValidTimes(),
        ) -> VCMintResponse:
            self.add_to_log(
                "vc_mint", (request.did_id, tx_config, request.target_address, request.fee, request.push, timelock_info)
            )

            return VCMintResponse(
                [STD_UTX],
                [STD_TX],
                VCRecord(
                    VerifiedCredential(
                        STD_TX.removals[0],
                        LineageProof(None, None, None),
                        VCLineageProof(None, None, None, None),
                        bytes32([3] * 32),
                        bytes32.zeros,
                        bytes32([1] * 32),
                        None,
                    ),
                    uint32(0),
                ),
            )

    inst_rpc_client = VcsMintRpcClient()
    test_rpc_clients.wallet_rpc_client = inst_rpc_client
    did_bytes = get_bytes32(1)
    did_id = encode_puzzle_hash(did_bytes, "did:chik:")
    target_bytes = get_bytes32(2)
    target_addr = encode_puzzle_hash(target_bytes, "xck")
    command_args = [
        "wallet",
        "vcs",
        "mint",
        FINGERPRINT_ARG,
        f"-d{did_id}",
        "-m0.5",
        f"-t{target_addr}",
        "--valid-at",
        "100",
        "--expires-at",
        "150",
    ]
    # these are various things that should be in the output
    assert_list = [
        f"New VC with launcher ID minted: {get_bytes32(3).hex()}",
        f"Transaction {get_bytes32(2).hex()}",
    ]
    run_cli_command_and_assert(capsys, root_dir, command_args, assert_list)
    expected_calls: logType = {
        "vc_mint": [(did_id, DEFAULT_TX_CONFIG, target_addr, 500000000000, True, test_condition_valid_times)]
    }
    test_rpc_clients.wallet_rpc_client.check_log(expected_calls)


def test_vcs_get(capsys: object, get_test_cli_clients: tuple[TestRpcClients, Path]) -> None:
    test_rpc_clients, root_dir = get_test_cli_clients

    # set RPC Client
    class VcsGetRpcClient(TestWalletRpcClient):
        async def vc_get_list(self, request: VCGetList) -> VCGetListResponse:
            self.add_to_log("vc_get_list", (request.start, request.end))
            proofs = [VCProofWithHash(get_bytes32(1), VCProofsRPC([("proof here", "")]))]
            records = [
                VCRecordWithCoinID(
                    VerifiedCredential(
                        STD_TX.removals[0],
                        LineageProof(None, None, None),
                        VCLineageProof(None, None, None, None),
                        bytes32([3] * 32),
                        bytes32.zeros,
                        bytes32([1] * 32),
                        None,
                    ),
                    uint32(0),
                    bytes32.zeros,
                )
            ]
            return VCGetListResponse(records, proofs)

    inst_rpc_client = VcsGetRpcClient()
    test_rpc_clients.wallet_rpc_client = inst_rpc_client
    command_args = ["wallet", "vcs", "get", FINGERPRINT_ARG, "-s10", "-c10"]
    # these are various things that should be in the output
    assert_list = [
        f"Proofs:\n- {get_bytes32(1).hex()}\n  - proof here",
        f"Launcher ID: {get_bytes32(3).hex()}",
        f"Inner Address: {encode_puzzle_hash(bytes32.zeros, 'xck')}",
    ]
    run_cli_command_and_assert(capsys, root_dir, command_args, assert_list)
    expected_calls: logType = {"vc_get_list": [(10, 10)]}
    test_rpc_clients.wallet_rpc_client.check_log(expected_calls)


def test_vcs_update_proofs(capsys: object, get_test_cli_clients: tuple[TestRpcClients, Path]) -> None:
    test_rpc_clients, root_dir = get_test_cli_clients

    # set RPC Client
    class VcsUpdateProofsRpcClient(TestWalletRpcClient):
        async def vc_spend(
            self,
            request: VCSpend,
            tx_config: TXConfig,
            timelock_info: ConditionValidTimes = ConditionValidTimes(),
        ) -> VCSpendResponse:
            self.add_to_log(
                "vc_spend",
                (
                    request.vc_id,
                    tx_config,
                    request.new_puzhash,
                    request.new_proof_hash,
                    request.provider_inner_puzhash,
                    request.fee,
                    request.push,
                    timelock_info,
                ),
            )
            return VCSpendResponse([STD_UTX], [STD_TX])

    inst_rpc_client = VcsUpdateProofsRpcClient()
    test_rpc_clients.wallet_rpc_client = inst_rpc_client
    vc_bytes = get_bytes32(1)
    target_ph = get_bytes32(2)
    new_proof = get_bytes32(3)
    command_args = [
        "wallet",
        "vcs",
        "update_proofs",
        FINGERPRINT_ARG,
        f"-l{vc_bytes.hex()}",
        "-m0.5",
        f"-t{target_ph.hex()}",
        f"-p{new_proof.hex()}",
        "--reuse-puzhash",
        "--valid-at",
        "100",
        "--expires-at",
        "150",
    ]
    # these are various things that should be in the output
    assert_list = [
        "Proofs successfully updated!",
        f"Transaction {get_bytes32(2).hex()}",
    ]
    run_cli_command_and_assert(capsys, root_dir, command_args, assert_list)
    expected_calls: logType = {
        "vc_spend": [
            (
                vc_bytes,
                DEFAULT_TX_CONFIG.override(reuse_puzhash=True),
                target_ph,
                new_proof,
                None,
                uint64(500000000000),
                True,
                test_condition_valid_times,
            )
        ]
    }
    test_rpc_clients.wallet_rpc_client.check_log(expected_calls)


def test_vcs_add_proof_reveal(capsys: object, get_test_cli_clients: tuple[TestRpcClients, Path]) -> None:
    test_rpc_clients, root_dir = get_test_cli_clients

    # set RPC Client
    class VcsAddProofRevealRpcClient(TestWalletRpcClient):
        async def vc_add_proofs(self, request: VCAddProofs) -> None:
            self.add_to_log("vc_add_proofs", (request.to_json_dict()["proofs"],))

    inst_rpc_client = VcsAddProofRevealRpcClient()
    test_rpc_clients.wallet_rpc_client = inst_rpc_client
    new_proof = "test_proof"
    command_args = ["wallet", "vcs", "add_proof_reveal", FINGERPRINT_ARG, f"-p{new_proof}"]
    # these are various things that should be in the output
    assert_list = ["Proofs added to DB successfully!"]
    run_cli_command_and_assert(capsys, root_dir, command_args, assert_list)

    root_assert_list = ["Proof Hash: 5fdf0dfd1fddc56c0f9f68fdb31390721625321ce79f5606b3d2c6ddebbf2a79"]
    run_cli_command_and_assert(capsys, root_dir, [*command_args, "-r"], root_assert_list)

    expected_calls: logType = {"vc_add_proofs": [({new_proof: "1"},)]}
    test_rpc_clients.wallet_rpc_client.check_log(expected_calls)


def test_vcs_get_proofs_for_root(capsys: object, get_test_cli_clients: tuple[TestRpcClients, Path]) -> None:
    test_rpc_clients, root_dir = get_test_cli_clients

    # set RPC Client
    class VcsGetProofsForRootRpcClient(TestWalletRpcClient):
        async def vc_get_proofs_for_root(self, request: VCGetProofsForRoot) -> VCGetProofsForRootResponse:
            self.add_to_log("vc_get_proofs_for_root", (request.root,))
            return VCGetProofsForRootResponse([("test_proof", "1"), ("test_proof2", "1")])

    inst_rpc_client = VcsGetProofsForRootRpcClient()
    test_rpc_clients.wallet_rpc_client = inst_rpc_client
    proof_hash = get_bytes32(1)
    command_args = ["wallet", "vcs", "get_proofs_for_root", FINGERPRINT_ARG, f"-r{proof_hash.hex()}"]
    # these are various things that should be in the output
    assert_list = ["Proofs:", "test_proof", "test_proof2"]
    run_cli_command_and_assert(capsys, root_dir, command_args, assert_list)

    expected_calls: logType = {"vc_get_proofs_for_root": [(proof_hash,)]}
    test_rpc_clients.wallet_rpc_client.check_log(expected_calls)


def test_vcs_revoke(capsys: object, get_test_cli_clients: tuple[TestRpcClients, Path]) -> None:
    test_rpc_clients, root_dir = get_test_cli_clients

    # set RPC Client
    class VcsRevokeRpcClient(TestWalletRpcClient):
        async def vc_get(self, request: VCGet) -> VCGetResponse:
            self.add_to_log("vc_get", (request.vc_id,))

            return VCGetResponse(
                VCRecord(
                    VerifiedCredential(
                        Coin(get_bytes32(1), get_bytes32(2), uint64(12345678)),
                        LineageProof(),
                        VCLineageProof(),
                        bytes32.zeros,
                        bytes32.zeros,
                        bytes32.zeros,
                        None,
                    ),
                    uint32(0),
                )
            )

        async def vc_revoke(
            self,
            request: VCRevoke,
            tx_config: TXConfig,
            timelock_info: ConditionValidTimes = ConditionValidTimes(),
        ) -> VCRevokeResponse:
            self.add_to_log("vc_revoke", (request.vc_parent_id, tx_config, request.fee, request.push, timelock_info))
            return VCRevokeResponse([STD_UTX], [STD_TX])

    inst_rpc_client = VcsRevokeRpcClient()
    test_rpc_clients.wallet_rpc_client = inst_rpc_client
    parent_id = get_bytes32(1)
    vc_id = get_bytes32(2)
    command_args = [
        "wallet",
        "vcs",
        "revoke",
        FINGERPRINT_ARG,
        "-m0.5",
        "--reuse-puzhash",
        "--valid-at",
        "100",
        "--expires-at",
        "150",
    ]
    # these are various things that should be in the output
    assert_list = ["VC successfully revoked!", f"Transaction {get_bytes32(2).hex()}"]
    run_cli_command_and_assert(capsys, root_dir, [*command_args, f"-p{parent_id.hex()}"], assert_list)
    run_cli_command_and_assert(capsys, root_dir, [*command_args, f"-l{vc_id.hex()}"], assert_list)
    expected_calls: logType = {
        "vc_get": [(vc_id,)],
        "vc_revoke": [
            (
                parent_id,
                DEFAULT_TX_CONFIG.override(reuse_puzhash=True),
                uint64(500000000000),
                True,
                test_condition_valid_times,
            ),
            (
                parent_id,
                DEFAULT_TX_CONFIG.override(reuse_puzhash=True),
                uint64(500000000000),
                True,
                test_condition_valid_times,
            ),
        ],
    }
    test_rpc_clients.wallet_rpc_client.check_log(expected_calls)


def test_vcs_approve_r_cats(capsys: object, get_test_cli_clients: tuple[TestRpcClients, Path]) -> None:
    test_rpc_clients, root_dir = get_test_cli_clients

    # set RPC Client
    class VcsApproveRCATSRpcClient(TestWalletRpcClient):
        async def crcat_approve_pending(
            self,
            wallet_id: uint32,
            min_amount_to_claim: uint64,
            tx_config: TXConfig,
            fee: uint64 = uint64(0),
            push: bool = True,
            timelock_info: ConditionValidTimes = ConditionValidTimes(),
        ) -> list[TransactionRecord]:
            self.add_to_log(
                "crcat_approve_pending",
                (
                    wallet_id,
                    min_amount_to_claim,
                    tx_config,
                    fee,
                    push,
                    timelock_info,
                ),
            )
            return [STD_TX]

    inst_rpc_client = VcsApproveRCATSRpcClient()
    test_rpc_clients.wallet_rpc_client = inst_rpc_client
    wallet_id = uint32(2)
    command_args = [
        "wallet",
        "vcs",
        "approve_r_cats",
        FINGERPRINT_ARG,
        f"-i{wallet_id}",
        "-a1",
        "-m0.5",
        "--min-coin-amount",
        "0.001",
        "--max-coin-amount",
        "10",
        "--reuse",
        "--valid-at",
        "100",
        "--expires-at",
        "150",
    ]
    # these are various things that should be in the output
    assert_list = ["VC successfully approved R-CATs!", f"Transaction {get_bytes32(2).hex()}"]
    run_cli_command_and_assert(capsys, root_dir, command_args, assert_list)
    expected_calls: logType = {
        "crcat_approve_pending": [
            (
                wallet_id,
                uint64(1000),
                TXConfig(
                    min_coin_amount=uint64(1),
                    max_coin_amount=uint64(10000),
                    excluded_coin_amounts=[],
                    excluded_coin_ids=[],
                    reuse_puzhash=True,
                ),
                uint64(500000000000),
                True,
                test_condition_valid_times,
            )
        ],
        "get_wallets": [(None,)],
    }
    test_rpc_clients.wallet_rpc_client.check_log(expected_calls)



from __future__ import annotations

from typing import Any

from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.condition_opcodes import ConditionOpcode
from chik.util.ints import uint64
from chik.util.streamable import VersionedBlob
from chik.wallet.payment import Payment
from chik.wallet.puzzles.clawback.drivers import create_merkle_puzzle
from chik.wallet.puzzles.clawback.metadata import ClawbackMetadata, ClawbackVersion
from chik.wallet.util.wallet_types import RemarkDataType


class ClawbackPuzzleDecorator:
    """
    This class is a wrapper for clawback puzzles. It allows us to add Clawback characteristics to the inner puzzle.
    """

    time_lock: uint64

    @staticmethod
    def create(config: dict[str, Any]) -> ClawbackPuzzleDecorator:
        self = ClawbackPuzzleDecorator()
        self.time_lock = uint64(config.get("clawback_timelock", 0))
        return self

    def decorate(self, inner_puzzle: Program) -> Program:
        # We don't wrap anything for the Clawback
        return inner_puzzle

    def decorate_target_puzzle_hash(
        self,
        inner_puzzle: Program,
        target_puzzle_hash: bytes32,
    ) -> tuple[Program, bytes32]:
        return (
            self.decorate(inner_puzzle),
            create_merkle_puzzle(self.time_lock, inner_puzzle.get_tree_hash(), target_puzzle_hash).get_tree_hash(),
        )

    def solve(
        self, inner_puzzle: Program, primaries: list[Payment], inner_solution: Program
    ) -> tuple[Program, Program]:
        # Append REMARK condition [1, "CLAWBACK", TIME_LOCK, SENDER_PUZHSAH, RECIPIENT_PUZHSAH]
        if len(primaries) == 1:
            recipient_puzhash = primaries[0].puzzle_hash
            metadata = ClawbackMetadata(self.time_lock, inner_puzzle.get_tree_hash(), recipient_puzhash)
            remark_condition = Program.to(
                [
                    ConditionOpcode.REMARK.value,
                    RemarkDataType.CLAWBACK,
                    bytes(VersionedBlob(ClawbackVersion.V1.value, bytes(metadata))),
                ]
            )
            # Insert the REMARK condition into the condition list
            conditions = remark_condition.cons(inner_solution.rest().first().rest())
            conditions = inner_solution.rest().first().first().cons(conditions)
            inner_solution = inner_solution.replace(rf=conditions)
        return self.decorate(inner_puzzle), inner_solution

    def decorate_memos(
        self, inner_puzzle: Program, target_puzzle_hash: bytes32, memos: list[bytes]
    ) -> tuple[Program, list[bytes]]:
        memos.insert(0, target_puzzle_hash)
        return self.decorate(inner_puzzle), memos



from __future__ import annotations

import logging
import time
from random import randrange

import pytest

from chik.consensus.default_constants import DEFAULT_CONSTANTS
from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.hash import std_hash
from chik.util.ints import uint32, uint64, uint128
from chik.wallet.coin_selection import (
    check_for_exact_match,
    knapsack_coin_algorithm,
    select_coins,
    select_smallest_coin_over_target,
    sum_largest_coins,
)
from chik.wallet.util.tx_config import DEFAULT_COIN_SELECTION_CONFIG
from chik.wallet.util.wallet_types import WalletType
from chik.wallet.wallet_coin_record import WalletCoinRecord

log = logging.getLogger(__name__)


class TestCoinSelection:
    @pytest.fixture(scope="function")
    def a_hash(self) -> bytes32:
        return std_hash(b"a")

    def test_exact_match(self, a_hash: bytes32) -> None:
        coin_list = [
            Coin(a_hash, a_hash, uint64(220000)),
            Coin(a_hash, a_hash, uint64(120000)),
            Coin(a_hash, a_hash, uint64(22)),
        ]
        assert check_for_exact_match(coin_list, uint64(220000)) == coin_list[0]
        assert check_for_exact_match(coin_list, uint64(22)) == coin_list[2]
        # check for no match.
        assert check_for_exact_match(coin_list, uint64(20)) is None

    def test_knapsack_coin_selection(self, a_hash: bytes32) -> None:
        tries = 100
        coins_to_append = 1000
        amounts = list(range(1, coins_to_append))
        amounts.sort(reverse=True)
        coin_list: list[Coin] = [Coin(a_hash, a_hash, uint64(100000000 * a)) for a in amounts]
        for i in range(tries):
            knapsack = knapsack_coin_algorithm(
                coin_list, uint128(30000000000000), DEFAULT_CONSTANTS.MAX_COIN_AMOUNT, 999999, seed=bytes([i])
            )
            assert knapsack is not None
            assert sum(coin.amount for coin in knapsack) >= 310000000

    def test_knapsack_coin_selection_2(self, a_hash: bytes32) -> None:
        coin_amounts = [6, 20, 40, 80, 150, 160, 203, 202, 201, 320]
        coin_amounts.sort(reverse=True)
        coin_list: list[Coin] = [Coin(a_hash, a_hash, uint64(a)) for a in coin_amounts]
        # coin_list = set([coin for a in coin_amounts])
        for i in range(100):
            knapsack = knapsack_coin_algorithm(
                coin_list, uint128(265), DEFAULT_CONSTANTS.MAX_COIN_AMOUNT, 99999, seed=bytes([i])
            )
            assert knapsack is not None
            selected_sum = sum(coin.amount for coin in list(knapsack))
            assert 265 <= selected_sum <= 281  # Selects a set of coins which does exceed by too much

    @pytest.mark.anyio
    async def test_coin_selection_randomly(self, a_hash: bytes32) -> None:
        coin_base_amounts = [3, 6, 20, 40, 80, 150, 160, 203, 202, 201, 320]
        coin_amounts = []
        spendable_amount = 0
        # this is possibly overkill, but it's a good test.
        for i in range(3000):
            for amount in coin_base_amounts:
                c_amount = randrange(1, 10000000) * amount
                coin_amounts.append(c_amount)
                spendable_amount += c_amount
        spendable_amount = uint128(spendable_amount)

        coin_list: list[WalletCoinRecord] = [
            WalletCoinRecord(Coin(a_hash, a_hash, uint64(a)), uint32(1), uint32(1), False, True, WalletType(0), 1)
            for a in coin_amounts
        ]
        for target_amount in coin_amounts[:100]:  # select the first 100 values
            result: set[Coin] = await select_coins(
                spendable_amount,
                DEFAULT_COIN_SELECTION_CONFIG,
                coin_list,
                {},
                logging.getLogger("test"),
                uint128(target_amount),
            )
            assert result is not None
            assert sum(coin.amount for coin in result) >= target_amount
            assert len(result) <= 500

    @pytest.mark.anyio
    async def test_coin_selection_with_dust(self, a_hash: bytes32) -> None:
        spendable_amount = uint128(5000000000000 + 10000)
        coin_list: list[WalletCoinRecord] = [
            WalletCoinRecord(
                Coin(a_hash, a_hash, uint64(5000000000000)), uint32(1), uint32(1), False, True, WalletType(0), 1
            )
        ]
        for i in range(10000):
            coin_list.append(
                WalletCoinRecord(
                    Coin(a_hash, std_hash(i.to_bytes(length=32, byteorder="big")), uint64(1)),
                    uint32(1),
                    uint32(1),
                    False,
                    True,
                    WalletType(0),
                    1,
                )
            )
        # make sure coins are not identical.
        for target_amount in [10000, 9999]:
            print("Target amount: ", target_amount)
            result: set[Coin] = await select_coins(
                spendable_amount,
                DEFAULT_COIN_SELECTION_CONFIG,
                coin_list,
                {},
                logging.getLogger("test"),
                uint128(target_amount),
            )
            assert result is not None
            assert sum(coin.amount for coin in result) >= target_amount
            assert len(result) == 1  # only one coin should be selected

        for i in range(100):
            coin_list.append(
                WalletCoinRecord(
                    Coin(a_hash, std_hash(i.to_bytes(length=32, byteorder="big")), uint64(2000)),
                    uint32(1),
                    uint32(1),
                    False,
                    True,
                    WalletType(0),
                    1,
                )
            )
        spendable_amount = uint128(spendable_amount + 2000 * 100)
        for target_amount in [50000, 25000, 15000, 10000, 9000, 3000]:  # select the first 100 values
            dusty_result: set[Coin] = await select_coins(
                spendable_amount,
                DEFAULT_COIN_SELECTION_CONFIG,
                coin_list,
                {},
                logging.getLogger("test"),
                uint128(target_amount),
            )
            assert dusty_result is not None
            assert sum(coin.amount for coin in dusty_result) >= target_amount
            for coin in dusty_result:
                assert coin.amount > 1
            assert len(dusty_result) <= 500

        # test when we have multiple coins under target, and a lot of dust coins.
        spendable_amount = uint128(25000 + 10000)
        new_coin_list: list[WalletCoinRecord] = []
        for i in range(5):
            new_coin_list.append(
                WalletCoinRecord(
                    Coin(a_hash, std_hash(i.to_bytes(length=32, byteorder="big")), uint64(5000)),
                    uint32(1),
                    uint32(1),
                    False,
                    True,
                    WalletType(0),
                    1,
                )
            )

        for i in range(10000):
            new_coin_list.append(
                WalletCoinRecord(
                    Coin(a_hash, std_hash(i.to_bytes(length=32, byteorder="big")), uint64(1)),
                    uint32(1),
                    uint32(1),
                    False,
                    True,
                    WalletType(0),
                    1,
                )
            )
        for target_amount in [20000, 15000, 10000, 5000]:  # select the first 100 values
            dusty_below_target: set[Coin] = await select_coins(
                spendable_amount,
                DEFAULT_COIN_SELECTION_CONFIG,
                new_coin_list,
                {},
                logging.getLogger("test"),
                uint128(target_amount),
            )
            assert dusty_below_target is not None
            assert sum(coin.amount for coin in dusty_below_target) >= target_amount
            for coin in dusty_below_target:
                assert coin.amount == 5000
            assert len(dusty_below_target) <= 500

    @pytest.mark.anyio
    async def test_dust_and_one_large_coin(self, a_hash: bytes32) -> None:
        # test when we have a lot of dust and 1 large coin
        spendable_amount = uint128(50000 + 10000)
        new_coin_list: list[WalletCoinRecord] = [
            WalletCoinRecord(
                Coin(a_hash, std_hash(b"123"), uint64(50000)), uint32(1), uint32(1), False, True, WalletType(0), 1
            )
        ]

        for i in range(10000):
            new_coin_list.append(
                WalletCoinRecord(
                    Coin(a_hash, std_hash(i.to_bytes(length=32, byteorder="big")), uint64(1)),
                    uint32(1),
                    uint32(1),
                    False,
                    True,
                    WalletType(0),
                    1,
                )
            )
        for target_amount in [50000, 10001, 10000, 9999]:
            dusty_below_target: set[Coin] = await select_coins(
                spendable_amount,
                DEFAULT_COIN_SELECTION_CONFIG,
                new_coin_list,
                {},
                logging.getLogger("test"),
                uint128(target_amount),
            )
            assert dusty_below_target is not None
            assert sum(coin.amount for coin in dusty_below_target) >= target_amount
            assert len(dusty_below_target) <= 500

    @pytest.mark.anyio
    async def test_coin_selection_failure(self, a_hash: bytes32) -> None:
        spendable_amount = uint128(10000)
        coin_list: list[WalletCoinRecord] = []
        for i in range(10000):
            coin_list.append(
                WalletCoinRecord(
                    Coin(a_hash, std_hash(i.to_bytes(length=32, byteorder="big")), uint64(1)),
                    uint32(1),
                    uint32(1),
                    False,
                    True,
                    WalletType(0),
                    1,
                )
            )
        # make sure coins are not identical.
        # test for failure
        with pytest.raises(ValueError):
            for target_amount in [10000, 9999]:
                await select_coins(
                    spendable_amount,
                    DEFAULT_COIN_SELECTION_CONFIG,
                    coin_list,
                    {},
                    logging.getLogger("test"),
                    uint128(target_amount),
                )
        # test not enough coin failure.
        with pytest.raises(ValueError):
            for target_amount in [10001, 20000]:
                await select_coins(
                    spendable_amount,
                    DEFAULT_COIN_SELECTION_CONFIG,
                    coin_list,
                    {},
                    logging.getLogger("test"),
                    uint128(target_amount),
                )

    @pytest.mark.anyio
    async def test_coin_selection(self, a_hash: bytes32) -> None:
        coin_amounts = [3, 6, 20, 40, 80, 150, 160, 203, 202, 201, 320]
        coin_list: list[WalletCoinRecord] = [
            WalletCoinRecord(Coin(a_hash, a_hash, uint64(a)), uint32(1), uint32(1), False, True, WalletType(0), 1)
            for a in coin_amounts
        ]
        spendable_amount = uint128(sum(coin_amounts))

        # check for exact match
        target_amount = uint128(40)
        exact_match_result: set[Coin] = await select_coins(
            spendable_amount,
            DEFAULT_COIN_SELECTION_CONFIG,
            coin_list,
            {},
            logging.getLogger("test"),
            target_amount,
        )
        assert exact_match_result is not None
        assert sum(coin.amount for coin in exact_match_result) >= target_amount
        assert len(exact_match_result) == 1

        # check for match of 2
        target_amount = uint128(153)
        match_2: set[Coin] = await select_coins(
            spendable_amount,
            DEFAULT_COIN_SELECTION_CONFIG,
            coin_list,
            {},
            logging.getLogger("test"),
            target_amount,
        )
        assert match_2 is not None
        assert sum(coin.amount for coin in match_2) == target_amount
        assert len(match_2) == 2
        # check for match of at least 3. it is random after all.
        target_amount = uint128(541)
        match_3: set[Coin] = await select_coins(
            spendable_amount,
            DEFAULT_COIN_SELECTION_CONFIG,
            coin_list,
            {},
            logging.getLogger("test"),
            target_amount,
        )
        assert match_3 is not None
        assert sum(coin.amount for coin in match_3) >= target_amount
        assert len(match_3) >= 3

        # check for match of all
        target_amount = spendable_amount
        match_all: set[Coin] = await select_coins(
            spendable_amount,
            DEFAULT_COIN_SELECTION_CONFIG,
            coin_list,
            {},
            logging.getLogger("test"),
            target_amount,
        )
        assert match_all is not None
        assert sum(coin.amount for coin in match_all) == target_amount
        assert len(match_all) == len(coin_list)

        # test smallest greater than target
        greater_coin_amounts = [1, 2, 5, 20, 400, 700]
        greater_coin_list: list[WalletCoinRecord] = [
            WalletCoinRecord(Coin(a_hash, a_hash, uint64(a)), uint32(1), uint32(1), False, True, WalletType(0), 1)
            for a in greater_coin_amounts
        ]
        greater_spendable_amount = uint128(sum(greater_coin_amounts))
        target_amount = uint128(625)
        smallest_result: set[Coin] = await select_coins(
            greater_spendable_amount,
            DEFAULT_COIN_SELECTION_CONFIG,
            greater_coin_list,
            {},
            logging.getLogger("test"),
            target_amount,
        )
        assert smallest_result is not None
        assert sum(coin.amount for coin in smallest_result) > target_amount
        assert len(smallest_result) == 1

        # test smallest greater than target with only 1 large coin.
        single_greater_coin_list: list[WalletCoinRecord] = [
            WalletCoinRecord(Coin(a_hash, a_hash, uint64(70000)), uint32(1), uint32(1), False, True, WalletType(0), 1)
        ]
        single_greater_spendable_amount = uint128(70000)
        target_amount = uint128(50000)
        single_greater_result: set[Coin] = await select_coins(
            single_greater_spendable_amount,
            DEFAULT_COIN_SELECTION_CONFIG,
            single_greater_coin_list,
            {},
            logging.getLogger("test"),
            target_amount,
        )
        assert single_greater_result is not None
        assert sum(coin.amount for coin in single_greater_result) > target_amount
        assert len(single_greater_result) == 1

        # test smallest greater than target with only multiple larger then target coins.
        multiple_greater_coin_amounts = [90000, 100000, 120000, 200000, 100000]
        multiple_greater_coin_list: list[WalletCoinRecord] = [
            WalletCoinRecord(Coin(a_hash, a_hash, uint64(a)), uint32(1), uint32(1), False, True, WalletType(0), 1)
            for a in multiple_greater_coin_amounts
        ]
        multiple_greater_spendable_amount = uint128(sum(multiple_greater_coin_amounts))
        target_amount = uint128(70000)
        multiple_greater_result: set[Coin] = await select_coins(
            multiple_greater_spendable_amount,
            DEFAULT_COIN_SELECTION_CONFIG,
            multiple_greater_coin_list,
            {},
            logging.getLogger("test"),
            target_amount,
        )
        assert multiple_greater_result is not None
        assert sum(coin.amount for coin in multiple_greater_result) > target_amount
        assert sum(coin.amount for coin in multiple_greater_result) == 90000
        assert len(multiple_greater_result) == 1

    @pytest.mark.anyio
    async def test_coin_selection_difficult(self, a_hash: bytes32) -> None:
        num_coins = 40
        spendable_amount = uint128(num_coins * 1000)
        coin_list: list[WalletCoinRecord] = [
            WalletCoinRecord(
                Coin(a_hash, std_hash(i.to_bytes(4, "big")), uint64(1000)),
                uint32(1),
                uint32(1),
                False,
                True,
                WalletType(0),
                1,
            )
            for i in range(num_coins)
        ]
        target_amount = spendable_amount - 1
        result: set[Coin] = await select_coins(
            spendable_amount,
            DEFAULT_COIN_SELECTION_CONFIG,
            coin_list,
            {},
            logging.getLogger("test"),
            uint128(target_amount),
        )
        assert result is not None
        print(result)
        print(sum(c.amount for c in result))
        assert sum(coin.amount for coin in result) >= target_amount

    @pytest.mark.anyio
    async def test_smallest_coin_over_amount(self, a_hash: bytes32) -> None:
        coin_list: list[Coin] = [
            Coin(a_hash, std_hash(i.to_bytes(4, "big")), uint64((39 - i) * 1000)) for i in range(40)
        ]
        assert select_smallest_coin_over_target(uint128(100), coin_list) == coin_list[39 - 1]
        assert select_smallest_coin_over_target(uint128(1000), coin_list) == coin_list[39 - 1]
        assert select_smallest_coin_over_target(uint128(1001), coin_list) == coin_list[39 - 2]
        assert select_smallest_coin_over_target(uint128(37000), coin_list) == coin_list[39 - 37]
        assert select_smallest_coin_over_target(uint128(39000), coin_list) == coin_list[39 - 39]
        assert select_smallest_coin_over_target(uint128(39001), coin_list) is None

    @pytest.mark.anyio
    async def test_sum_largest_coins(self, a_hash: bytes32) -> None:
        coin_list: list[Coin] = list(
            reversed([Coin(a_hash, std_hash(i.to_bytes(4, "big")), uint64(i)) for i in range(41)])
        )
        assert sum_largest_coins(uint128(40), coin_list) == {coin_list[0]}
        assert sum_largest_coins(uint128(79), coin_list) == {coin_list[0], coin_list[1]}
        assert sum_largest_coins(uint128(40000), coin_list) is None

    @pytest.mark.anyio
    async def test_knapsack_perf(self, a_hash: bytes32) -> None:
        start = time.time()
        coin_list: list[Coin] = [
            Coin(a_hash, std_hash(i.to_bytes(4, "big")), uint64((200000 - i) * 1000)) for i in range(200000)
        ]
        knapsack_coin_algorithm(coin_list, uint128(2000000), 9999999999999999, 500)

        # Just a sanity check, it's actually much faster than this time
        assert time.time() - start < 10000

    @pytest.mark.anyio
    async def test_coin_selection_min_coin(self, a_hash: bytes32) -> None:
        spendable_amount = uint128(5000000 + 500 + 40050)
        coin_list: list[WalletCoinRecord] = [
            WalletCoinRecord(Coin(a_hash, a_hash, uint64(5000000)), uint32(1), uint32(1), False, True, WalletType(0), 1)
        ]
        for i in range(500):
            coin_list.append(
                WalletCoinRecord(
                    Coin(a_hash, std_hash(i.to_bytes(length=32, byteorder="big")), uint64(1)),
                    uint32(1),
                    uint32(1),
                    False,
                    True,
                    WalletType(0),
                    1,
                )
            )
        for i in range(1, 90):
            coin_list.append(
                WalletCoinRecord(
                    Coin(a_hash, std_hash(i.to_bytes(length=32, byteorder="big")), uint64(i * 10)),
                    uint32(1),
                    uint32(1),
                    False,
                    True,
                    WalletType(0),
                    1,
                )
            )
        # make sure coins are not identical.
        for target_amount in [500, 1000, 50000, 500000]:
            for min_coin_amount in [10, 100, 200, 300, 1000]:
                result: set[Coin] = await select_coins(
                    spendable_amount,
                    DEFAULT_COIN_SELECTION_CONFIG.override(min_coin_amount=uint64(min_coin_amount)),
                    coin_list,
                    {},
                    logging.getLogger("test"),
                    uint128(target_amount),
                )
                assert result is not None  # this should never happen
                assert sum(coin.amount for coin in result) >= target_amount
                for coin in result:
                    assert not coin.amount < min_coin_amount
                assert len(result) <= 500

    @pytest.mark.anyio
    async def test_coin_selection_with_excluded_coins(self) -> None:
        a_hash = std_hash(b"a")
        b_hash = std_hash(b"b")
        c_hash = std_hash(b"c")
        target_amount = uint128(2)
        spendable_coins = [
            Coin(a_hash, a_hash, uint64(3)),
            Coin(b_hash, b_hash, uint64(6)),
            Coin(c_hash, c_hash, uint64(9)),
        ]
        spendable_amount = uint128(sum(coin.amount for coin in spendable_coins))
        spendable_wallet_coin_records = [
            WalletCoinRecord(spendable_coin, uint32(1), uint32(1), False, True, WalletType(0), 1)
            for spendable_coin in spendable_coins
        ]
        excluded_coins = [Coin(a_hash, a_hash, uint64(3)), Coin(c_hash, c_hash, uint64(9))]
        # test that excluded coins are not included in the result
        selected_coins: set[Coin] = await select_coins(
            spendable_amount,
            DEFAULT_COIN_SELECTION_CONFIG.override(excluded_coin_ids=[c.name() for c in excluded_coins]),
            spendable_wallet_coin_records,
            {},
            logging.getLogger("test"),
            amount=target_amount,
        )

        assert selected_coins is not None
        assert sum(coin.amount for coin in selected_coins) >= target_amount
        assert len(selected_coins) == 1
        assert next(iter(selected_coins)) == Coin(b_hash, b_hash, uint64(6))

        excluded_all_coins = spendable_coins
        # make sure that a failure is raised if all coins are excluded.
        with pytest.raises(ValueError):
            await select_coins(
                spendable_amount,
                DEFAULT_COIN_SELECTION_CONFIG.override(excluded_coin_ids=[c.name() for c in excluded_all_coins]),
                spendable_wallet_coin_records,
                {},
                logging.getLogger("test"),
                amount=target_amount,
            )

    @pytest.mark.anyio
    async def test_coin_selection_with_zero_amount(self, a_hash: bytes32) -> None:
        coin_amounts = [3, 6, 20, 40, 80, 150, 160, 203, 202, 201, 320]
        coin_list: list[WalletCoinRecord] = [
            WalletCoinRecord(Coin(a_hash, a_hash, uint64(a)), uint32(1), uint32(1), False, True, WalletType(0), 1)
            for a in coin_amounts
        ]
        spendable_amount = uint128(sum(coin_amounts))

        # validate that a zero amount is handled correctly
        target_amount = uint128(0)
        zero_amount_result: set[Coin] = await select_coins(
            spendable_amount,
            DEFAULT_COIN_SELECTION_CONFIG,
            coin_list,
            {},
            logging.getLogger("test"),
            target_amount,
        )
        assert zero_amount_result is not None
        assert sum(coin.amount for coin in zero_amount_result) >= target_amount
        assert len(zero_amount_result) == 1
        # make sure that a failure is properly raised if we don't have any coins.
        with pytest.raises(ValueError):
            await select_coins(
                uint128(0),
                DEFAULT_COIN_SELECTION_CONFIG,
                [],
                {},
                logging.getLogger("test"),
                target_amount,
            )



from __future__ import annotations

from chik_rs import AugSchemeMPL
from klvm.operators import KEYWORD_FROM_ATOM
from klvm_tools.binutils import disassemble as bu_disassemble

from chik.consensus.default_constants import DEFAULT_CONSTANTS
from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.program import INFINITE_COST, Program
from chik.types.condition_opcodes import ConditionOpcode
from chik.util.condition_tools import conditions_dict_for_solution, pkm_pairs_for_conditions_dict
from chik.util.hash import std_hash
from chik.wallet.uncurried_puzzle import UncurriedPuzzle

CONDITIONS = {opcode.name: opcode.value[0] for opcode in ConditionOpcode}
KFA = {v: k for k, v in CONDITIONS.items()}


# information needed to spend a cc
# if we ever support more genesis conditions, like a re-issuable coin,
# we may need also to save the `genesis_coin_mod` or its hash


def disassemble(sexp: Program):
    """
    This version of `disassemble` also disassembles condition opcodes like `ASSERT_ANNOUNCEMENT_CONSUMED`.
    """
    kfa = dict(KEYWORD_FROM_ATOM)
    kfa.update((Program.to(k).as_atom(), v) for k, v in KFA.items())
    return bu_disassemble(sexp, kfa)


def coin_as_program(coin: Coin) -> Program:
    """
    Convenience function for when putting `coin_info` into a solution.
    """
    return Program.to([coin.parent_coin_info, coin.puzzle_hash, coin.amount])


def dump_coin(coin: Coin) -> str:
    return disassemble(coin_as_program(coin))


def recursive_uncurry_dump(puzzle: Program, layer: int, prefix: str, uncurried_already: UncurriedPuzzle) -> None:
    mod = uncurried_already.mod
    curried_args = uncurried_already.args
    if mod != puzzle:
        print(f"{prefix}- Layer {layer}:")
        print(f"{prefix}  - Mod hash: {mod.get_tree_hash().hex()}")
        for arg in curried_args.as_iter():
            uncurry_dump(arg, prefix=f"{prefix}  ")
        mod2, curried_args2 = mod.uncurry()
        if mod2 != mod:
            recursive_uncurry_dump(mod, layer + 1, prefix, UncurriedPuzzle(mod2, curried_args2))
    else:
        print(f"{prefix}- {bu_disassemble(puzzle)}")


def uncurry_dump(puzzle: Program, prefix: str = "") -> None:
    mod, curried_args = puzzle.uncurry()
    if mod != puzzle:
        print(f"{prefix}- <curried puzzle>")
        prefix = f"{prefix}  "

    recursive_uncurry_dump(puzzle, 1, prefix, UncurriedPuzzle(mod, curried_args))


def debug_spend_bundle(spend_bundle, agg_sig_additional_data=DEFAULT_CONSTANTS.AGG_SIG_ME_ADDITIONAL_DATA) -> None:
    """
    Print a lot of useful information about a `SpendBundle` that might help with debugging
    its klvm.
    """

    pks = []
    msgs = []

    created_coin_announcements: list[list[bytes]] = []
    asserted_coin_announcements = []
    created_puzzle_announcements: list[list[bytes]] = []
    asserted_puzzle_announcements = []

    print("=" * 80)
    for coin_spend in spend_bundle.coin_spends:
        coin = coin_spend.coin
        puzzle_reveal = Program.from_bytes(bytes(coin_spend.puzzle_reveal))
        solution = Program.from_bytes(bytes(coin_spend.solution))
        coin_name = coin.name()

        print(f"consuming coin {dump_coin(coin)}")
        print(f"  with id {coin_name.hex()}")
        print()
        print(f"\nbrun -y main.sym '{bu_disassemble(puzzle_reveal)}' '{bu_disassemble(solution)}'")

        print()
        print("--- Uncurried Args ---")
        uncurry_dump(puzzle_reveal)

        if puzzle_reveal.get_tree_hash() != coin_spend.coin.puzzle_hash:
            print()
            print("*** BAD PUZZLE REVEAL")
            print(f"{puzzle_reveal.get_tree_hash().hex()} vs {coin_spend.coin.puzzle_hash.hex()}")
            print("*" * 80)
            print()
            continue

        conditions = conditions_dict_for_solution(puzzle_reveal, solution, INFINITE_COST)
        for pk, m in pkm_pairs_for_conditions_dict(conditions, coin, agg_sig_additional_data):
            pks.append(pk)
            msgs.append(m)
        print()
        _cost, r = puzzle_reveal.run_with_cost(INFINITE_COST, solution)
        print(disassemble(r))
        create_coin_conditions = [con for con in r.as_iter() if con.first().as_int() == 51]
        print()
        if conditions and len(conditions) > 0:
            print("grouped conditions:")
            for condition_programs in conditions.values():
                print()
                for c in condition_programs:
                    if len(c.vars) == 0:
                        as_prog = Program.to([c.opcode])
                    elif len(c.vars) == 1:
                        as_prog = Program.to([c.opcode, c.vars[0]])
                    elif len(c.vars) == 2:
                        if c.opcode == ConditionOpcode.CREATE_COIN:
                            cc = next(
                                cc
                                for cc in create_coin_conditions
                                if cc.at("rf").atom == c.vars[0] and cc.at("rrf").atom == c.vars[1]
                            )
                            if cc.at("rrr").atom is None:
                                as_prog = Program.to([c.opcode, c.vars[0], c.vars[1], cc.at("rrrf")])
                            else:
                                as_prog = Program.to([c.opcode, c.vars[0], c.vars[1]])
                        else:
                            as_prog = Program.to([c.opcode, c.vars[0], c.vars[1]])
                    else:
                        raise Exception(f"Unexpected number of vars: {len(c.vars)}")

                    print(f"  {disassemble(as_prog)}")
            created_coin_announcements.extend(
                [coin_name, *_.vars] for _ in conditions.get(ConditionOpcode.CREATE_COIN_ANNOUNCEMENT, [])
            )
            asserted_coin_announcements.extend(
                [_.vars[0].hex() for _ in conditions.get(ConditionOpcode.ASSERT_COIN_ANNOUNCEMENT, [])]
            )
            created_puzzle_announcements.extend(
                [puzzle_reveal.get_tree_hash(), *_.vars]
                for _ in conditions.get(ConditionOpcode.CREATE_PUZZLE_ANNOUNCEMENT, [])
            )
            asserted_puzzle_announcements.extend(
                [_.vars[0].hex() for _ in conditions.get(ConditionOpcode.ASSERT_PUZZLE_ANNOUNCEMENT, [])]
            )
            print()
        else:
            print("(no output conditions generated)")
        print()
        print("-------")

    created = set(spend_bundle.additions())
    spent = set(spend_bundle.removals())

    zero_coin_set = {coin.name() for coin in created if coin.amount == 0}

    ephemeral = created.intersection(spent)
    created.difference_update(ephemeral)
    spent.difference_update(ephemeral)
    print()
    print("spent coins")
    for coin in sorted(spent, key=lambda _: _.name()):
        print(f"  {dump_coin(coin)}")
        print(f"      => spent coin id {coin.name().hex()}")
    print()
    print("created coins")
    for coin in sorted(created, key=lambda _: _.name()):
        print(f"  {dump_coin(coin)}")
        print(f"      => created coin id {coin.name().hex()}")

    if ephemeral:
        print()
        print("ephemeral coins")
        for coin in sorted(ephemeral, key=lambda _: _.name()):
            print(f"  {dump_coin(coin)}")
            print(f"      => created coin id {coin.name().hex()}")

    created_coin_announcement_pairs = [(_, std_hash(b"".join(_)).hex()) for _ in created_coin_announcements]
    if created_coin_announcement_pairs:
        print("created coin announcements")
        for announcement, hashed in sorted(created_coin_announcement_pairs, key=lambda _: _[-1]):
            as_hex = [f"0x{_.hex()}" for _ in announcement]
            print(f"  {as_hex} =>\n      {hashed}")

    eor_coin_announcements = sorted({_[-1] for _ in created_coin_announcement_pairs} ^ set(asserted_coin_announcements))

    created_puzzle_announcement_pairs = [(_, std_hash(b"".join(_)).hex()) for _ in created_puzzle_announcements]
    if created_puzzle_announcements:
        print("created puzzle announcements")
        for announcement, hashed in sorted(created_puzzle_announcement_pairs, key=lambda _: _[-1]):
            as_hex = [f"0x{_.hex()}" for _ in announcement]
            print(f"  {as_hex} =>\n      {hashed}")

    eor_puzzle_announcements = sorted(
        {_[-1] for _ in created_puzzle_announcement_pairs} ^ set(asserted_puzzle_announcements)
    )

    print()
    print()
    print(f"zero_coin_set = {sorted(zero_coin_set)}")
    print()
    if created_coin_announcement_pairs or asserted_coin_announcements:
        print(f"created  coin announcements = {sorted([_[-1] for _ in created_coin_announcement_pairs])}")
        print()
        print(f"asserted coin announcements = {sorted(asserted_coin_announcements)}")
        print()
        print(f"symdiff of coin announcements = {sorted(eor_coin_announcements)}")
        print()
    if created_puzzle_announcement_pairs or asserted_puzzle_announcements:
        print(f"created  puzzle announcements = {sorted([_[-1] for _ in created_puzzle_announcement_pairs])}")
        print()
        print(f"asserted puzzle announcements = {sorted(asserted_puzzle_announcements)}")
        print()
        print(f"symdiff of puzzle announcements = {sorted(eor_puzzle_announcements)}")
        print()
    print()
    print("=" * 80)
    print()
    validates = AugSchemeMPL.aggregate_verify(pks, msgs, spend_bundle.aggregated_signature)
    print(f"aggregated signature check pass: {validates}")
    print(f"pks: {pks}")
    print(f"msgs: {[msg.hex() for msg in msgs]}")
    print(f"add_data: {agg_sig_additional_data.hex()}")
    print(f"signature: {spend_bundle.aggregated_signature}")



from __future__ import annotations

from typing import Optional

from typing_extensions import Protocol

from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.wallet.puzzle_drivers import PuzzleInfo, Solver
from chik.wallet.uncurried_puzzle import UncurriedPuzzle


class DriverProtocol(Protocol):
    def match(self, puzzle: UncurriedPuzzle) -> Optional[PuzzleInfo]: ...

    def get_inner_puzzle(self, constructor: PuzzleInfo, puzzle_reveal: UncurriedPuzzle) -> Optional[Program]: ...

    def get_inner_solution(self, constructor: PuzzleInfo, solution: Program) -> Optional[Program]: ...

    def asset_id(self, constructor: PuzzleInfo) -> Optional[bytes32]: ...

    def construct(self, constructor: PuzzleInfo, inner_puzzle: Program) -> Program: ...

    def solve(
        self, constructor: PuzzleInfo, solver: Solver, inner_puzzle: Program, inner_solution: Program
    ) -> Program: ...



from __future__ import annotations

import dataclasses
from typing import Optional, TypeVar, Union

from aiosqlite import Row

from chik.data_layer.data_layer_wallet import Mirror, SingletonRecord
from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.db_wrapper import DBWrapper2, execute_fetchone
from chik.util.ints import uint16, uint32, uint64
from chik.wallet.lineage_proof import LineageProof

_T_DataLayerStore = TypeVar("_T_DataLayerStore", bound="DataLayerStore")


def _row_to_singleton_record(row: Row) -> SingletonRecord:
    return SingletonRecord(
        bytes32(row[0]),
        bytes32(row[1]),
        bytes32(row[2]),
        bytes32(row[3]),
        bool(row[4]),
        uint32(row[5]),
        LineageProof.from_bytes(row[6]),
        uint32(row[7]),
        uint64(row[8]),
    )


def _row_to_mirror(row: Row, confirmed_at_height: Optional[uint32]) -> Mirror:
    urls: list[bytes] = []
    byte_list: bytes = row[3]
    while byte_list != b"":
        length = uint16.from_bytes(byte_list[0:2])
        url = byte_list[2 : length + 2]
        byte_list = byte_list[length + 2 :]
        urls.append(url)
    return Mirror(bytes32(row[0]), bytes32(row[1]), uint64.from_bytes(row[2]), urls, bool(row[4]), confirmed_at_height)


class DataLayerStore:
    """
    WalletUserStore keeps track of all user created wallets and necessary smart-contract data
    """

    db_wrapper: DBWrapper2

    @classmethod
    async def create(cls: type[_T_DataLayerStore], db_wrapper: DBWrapper2) -> _T_DataLayerStore:
        self = cls()

        self.db_wrapper = db_wrapper

        async with self.db_wrapper.writer_maybe_transaction() as conn:
            await conn.execute(
                "CREATE TABLE IF NOT EXISTS singleton_records("
                "coin_id blob PRIMARY KEY,"
                " launcher_id blob,"
                " root blob,"
                " inner_puzzle_hash blob,"
                " confirmed tinyint,"
                " confirmed_at_height int,"
                " proof blob,"
                " generation int,"  # This first singleton will be 0, then 1, and so on.  This is handled by the DB.
                " timestamp int)"
            )

            await conn.execute(
                "CREATE TABLE IF NOT EXISTS mirrors("
                "coin_id blob PRIMARY KEY,"
                "launcher_id blob,"
                "amount blob,"
                "urls blob,"
                "ours tinyint)"
            )

            await conn.execute(
                "CREATE TABLE IF NOT EXISTS mirror_confirmations("
                "coin_id blob PRIMARY KEY,"
                "confirmed_at_height int)"
            )

            await conn.execute(
                "CREATE INDEX IF NOT EXISTS singleton_records_launcher_id_index ON singleton_records(launcher_id)"
            )
            await conn.execute("CREATE INDEX IF NOT EXISTS singleton_records_root_index ON singleton_records(root)")
            await conn.execute(
                "CREATE INDEX IF NOT EXISTS singleton_records_confirmed_at_height_index "
                "ON singleton_records(confirmed_at_height)"
            )
            await conn.execute(
                "CREATE INDEX IF NOT EXISTS singleton_records_generation_index ON singleton_records(generation)"
            )

            await conn.execute("CREATE TABLE IF NOT EXISTS launchers(id blob PRIMARY KEY, coin blob)")

            await conn.execute("CREATE INDEX IF NOT EXISTS mirrors_launcher_id_index ON mirrors(launcher_id)")

        return self

    async def _clear_database(self) -> None:
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            await (await conn.execute("DELETE FROM singleton_records")).close()

    async def add_singleton_record(self, record: SingletonRecord) -> None:
        """
        Store SingletonRecord in DB.
        """

        async with self.db_wrapper.writer_maybe_transaction() as conn:
            await conn.execute_insert(
                "INSERT OR REPLACE INTO singleton_records VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)",
                (
                    record.coin_id,
                    record.launcher_id,
                    record.root,
                    record.inner_puzzle_hash,
                    int(record.confirmed),
                    record.confirmed_at_height,
                    bytes(record.lineage_proof),
                    record.generation,
                    record.timestamp,
                ),
            )

    async def get_all_singletons_for_launcher(
        self,
        launcher_id: bytes32,
        min_generation: Optional[uint32] = None,
        max_generation: Optional[uint32] = None,
        num_results: Optional[uint32] = None,
    ) -> list[SingletonRecord]:
        """
        Returns stored singletons with a specific launcher ID.
        """
        query_params: list[Union[bytes32, uint32]] = [launcher_id]
        for optional_param in (min_generation, max_generation, num_results):
            if optional_param is not None:
                query_params.append(optional_param)

        async with self.db_wrapper.reader_no_transaction() as conn:
            cursor = await conn.execute(
                "SELECT * from singleton_records WHERE launcher_id=? "
                f"{'AND generation >=? ' if min_generation is not None else ''}"
                f"{'AND generation <=? ' if max_generation is not None else ''}"
                "ORDER BY generation DESC"
                f"{' LIMIT ?' if num_results is not None else ''}",
                tuple(query_params),
            )
            rows = await cursor.fetchall()
            await cursor.close()
        records = []

        for row in rows:
            records.append(_row_to_singleton_record(row))

        return records

    async def get_singleton_record(self, coin_id: bytes32) -> Optional[SingletonRecord]:
        """
        Checks DB for SingletonRecord with coin_id: coin_id and returns it.
        """
        # if tx_id in self.tx_record_cache:
        #     return self.tx_record_cache[tx_id]

        async with self.db_wrapper.reader_no_transaction() as conn:
            cursor = await conn.execute("SELECT * from singleton_records WHERE coin_id=?", (coin_id,))
            row = await cursor.fetchone()
            await cursor.close()
        if row is not None:
            return _row_to_singleton_record(row)
        return None

    async def get_latest_singleton(
        self, launcher_id: bytes32, only_confirmed: bool = False
    ) -> Optional[SingletonRecord]:
        """
        Checks DB for SingletonRecords with launcher_id: launcher_id and returns the most recent.
        """
        # if tx_id in self.tx_record_cache:
        #     return self.tx_record_cache[tx_id]
        async with self.db_wrapper.reader_no_transaction() as conn:
            if only_confirmed:
                # get latest confirmed root
                cursor = await conn.execute(
                    "SELECT * from singleton_records WHERE launcher_id=? and confirmed = 1 "
                    "ORDER BY generation DESC LIMIT 1",
                    (launcher_id,),
                )
            else:
                cursor = await conn.execute(
                    "SELECT * from singleton_records WHERE launcher_id=? ORDER BY generation DESC LIMIT 1",
                    (launcher_id,),
                )
            row = await cursor.fetchone()
            await cursor.close()
        if row is not None:
            return _row_to_singleton_record(row)
        return None

    async def get_unconfirmed_singletons(self, launcher_id: bytes32) -> list[SingletonRecord]:
        """
        Returns all singletons with a specific launcher id that have not yet been marked confirmed
        """
        async with self.db_wrapper.reader_no_transaction() as conn:
            cursor = await conn.execute(
                "SELECT * from singleton_records WHERE launcher_id=? AND confirmed=0", (launcher_id,)
            )
            rows = await cursor.fetchall()
            await cursor.close()
        records = [_row_to_singleton_record(row) for row in rows]

        return records

    async def get_singletons_by_root(self, launcher_id: bytes32, root: bytes32) -> list[SingletonRecord]:
        async with self.db_wrapper.reader_no_transaction() as conn:
            cursor = await conn.execute(
                "SELECT * from singleton_records WHERE launcher_id=? AND root=? ORDER BY generation DESC",
                (launcher_id, root),
            )
            rows = await cursor.fetchall()
            await cursor.close()
        records = []

        for row in rows:
            records.append(_row_to_singleton_record(row))

        return records

    async def set_confirmed(self, coin_id: bytes32, height: uint32, timestamp: uint64) -> None:
        """
        Updates singleton record to be confirmed.
        """
        current: Optional[SingletonRecord] = await self.get_singleton_record(coin_id)
        if current is None or current.confirmed_at_height == height:
            return

        await self.add_singleton_record(
            dataclasses.replace(current, confirmed=True, confirmed_at_height=height, timestamp=timestamp)
        )

    async def delete_singleton_record(self, coin_id: bytes32) -> None:
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            await (await conn.execute("DELETE FROM singleton_records WHERE coin_id=?", (coin_id,))).close()

    async def delete_singleton_records_by_launcher_id(self, launcher_id: bytes32) -> None:
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            await (await conn.execute("DELETE FROM singleton_records WHERE launcher_id=?", (launcher_id,))).close()

    async def add_launcher(self, launcher: Coin, confirmed_at_height: uint32) -> None:
        """
        Add a new launcher coin's information to the DB
        """
        launcher_bytes: bytes = (
            launcher.parent_coin_info + launcher.puzzle_hash + uint64(launcher.amount).stream_to_bytes()
        )
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            launcher_id = launcher.name()
            await conn.execute_insert(
                "INSERT OR REPLACE INTO launchers VALUES (?, ?)",
                (launcher_id, launcher_bytes),
            )

    async def get_launcher(self, launcher_id: bytes32) -> Optional[Coin]:
        """
        Checks DB for a launcher with the specified ID and returns it.
        """

        async with self.db_wrapper.reader_no_transaction() as conn:
            cursor = await conn.execute("SELECT * from launchers WHERE id=?", (launcher_id,))
            row = await cursor.fetchone()
            await cursor.close()
        if row is not None:
            return Coin(bytes32(row[1][0:32]), bytes32(row[1][32:64]), uint64(int.from_bytes(row[1][64:72], "big")))
        return None

    async def get_all_launchers(self) -> list[bytes32]:
        """
        Checks DB for all launchers.
        """

        async with self.db_wrapper.reader_no_transaction() as conn:
            cursor = await conn.execute("SELECT id from launchers")
            rows = await cursor.fetchall()
            await cursor.close()

        return [bytes32(row[0]) for row in rows]

    async def is_launcher_tracked(self, launcher_id: bytes32) -> bool:
        async with self.db_wrapper.reader_no_transaction() as conn:
            cursor = await conn.execute("SELECT COUNT(*) from singleton_records WHERE launcher_id=?", (launcher_id,))
            row = await cursor.fetchone()
            await cursor.close()
        if row is not None:
            count: int = row[0]
            return count > 0
        else:
            return False

    async def delete_launcher(self, launcher_id: bytes32) -> None:
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            await (await conn.execute("DELETE FROM launchers WHERE id=?", (launcher_id,))).close()

    async def add_mirror(self, mirror: Mirror) -> None:
        """
        Add a mirror coin to the DB
        """

        async with self.db_wrapper.writer_maybe_transaction() as conn:
            await conn.execute_insert(
                "INSERT OR REPLACE INTO mirrors VALUES (?, ?, ?, ?, ?)",
                (
                    mirror.coin_id,
                    mirror.launcher_id,
                    mirror.amount.stream_to_bytes(),
                    b"".join(
                        [uint16(len(url)).stream_to_bytes() + url for url in mirror.urls]
                    ),  # prefix each item with a length
                    1 if mirror.ours else 0,
                ),
            )
            await conn.execute_insert(
                "INSERT OR REPLACE INTO mirror_confirmations (coin_id, confirmed_at_height) VALUES (?, ?)",
                (
                    mirror.coin_id,
                    mirror.confirmed_at_height,
                ),
            )

    async def get_mirrors(self, launcher_id: bytes32) -> list[Mirror]:
        async with self.db_wrapper.reader_no_transaction() as conn:
            cursor = await conn.execute(
                "SELECT * from mirrors WHERE launcher_id=?",
                (launcher_id,),
            )
            rows = await cursor.fetchall()
            await cursor.close()
            mirrors: list[Mirror] = []

            for row in rows:
                confirmation_height = await execute_fetchone(
                    conn, "SELECT * FROM mirror_confirmations WHERE coin_id=?", (row[0],)
                )
                mirrors.append(
                    _row_to_mirror(row, None if confirmation_height is None else uint32(confirmation_height[1]))
                )

        return mirrors

    async def get_mirror(self, coin_id: bytes32) -> Mirror:
        async with self.db_wrapper.reader_no_transaction() as conn:
            cursor = await conn.execute(
                "SELECT * from mirrors WHERE coin_id=?",
                (coin_id,),
            )
            row = await cursor.fetchone()
            await cursor.close()
            assert row is not None
            confirmation_height = await execute_fetchone(
                conn, "SELECT * FROM mirror_confirmations WHERE coin_id=?", (row[0],)
            )

        return _row_to_mirror(row, None if confirmation_height is None else uint32(confirmation_height[1]))

    async def delete_mirror(self, coin_id: bytes32) -> None:
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            await (await conn.execute("DELETE FROM mirrors WHERE coin_id=?", (coin_id,))).close()
            await (await conn.execute("DELETE FROM mirror_confirmations WHERE coin_id=?", (coin_id,))).close()

    async def rollback_to_block(self, height: int) -> None:
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            cursor = await conn.execute(
                "SELECT * from mirror_confirmations WHERE confirmed_at_height>?",
                (height,),
            )
            rows = await cursor.fetchall()
            await cursor.close()

            for row in rows:
                await conn.execute(
                    "DELETE from mirror_confirmations WHERE coin_id=?",
                    (row[0],),
                )
                await conn.execute(
                    "DELETE from mirrors WHERE coin_id=?",
                    (row[0],),
                )

            cursor = await conn.execute(
                "SELECT launcher_id from singleton_records WHERE confirmed_at_height>? AND generation=0",
                (height,),
            )
            rows = await cursor.fetchall()
            await cursor.close()
            for row in rows:
                await conn.execute(
                    "DELETE from launchers WHERE id=?",
                    (row[0],),
                )

            await conn.execute(
                "UPDATE singleton_records SET "
                "confirmed_at_height = 0, confirmed = 0, timestamp = 0 WHERE confirmed_at_height > ?",
                (height,),
            )



from __future__ import annotations

from chik._tests.util.get_name_puzzle_conditions import get_name_puzzle_conditions
from chik.consensus.constants import ConsensusConstants
from chik.consensus.default_constants import DEFAULT_CONSTANTS
from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.full_block import FullBlock
from chik.types.generator_types import BlockGenerator
from chik.util.generator_tools import tx_removals_and_additions
from chik.util.ints import uint32


def run_and_get_removals_and_additions(
    block: FullBlock,
    max_cost: int,
    *,
    height: uint32,
    constants: ConsensusConstants = DEFAULT_CONSTANTS,
    mempool_mode: bool = False,
) -> tuple[list[bytes32], list[Coin]]:
    removals: list[bytes32] = []
    additions: list[Coin] = []

    assert len(block.transactions_generator_ref_list) == 0
    if not block.is_transaction_block():
        return [], []

    if block.transactions_generator is not None:
        npc_result = get_name_puzzle_conditions(
            BlockGenerator(block.transactions_generator, []),
            max_cost,
            mempool_mode=mempool_mode,
            height=height,
            constants=constants,
        )
        assert npc_result.error is None
        rem, add = tx_removals_and_additions(npc_result.conds)
        # build removals list
        removals.extend(rem)
        additions.extend(add)

    rewards = block.get_included_reward_coins()
    additions.extend(rewards)
    return removals, additions



from __future__ import annotations

import time
from dataclasses import dataclass
from ipaddress import IPv4Address, IPv6Address
from socket import AF_INET, AF_INET6, SOCK_STREAM
from typing import Optional, Union, cast
from unittest.mock import AsyncMock

import dns
import pytest
from dns.name import from_text
from dns.rdataclass import IN
from dns.rdatatype import AAAA as AAAA_TYPE
from dns.rdatatype import A as A_TYPE
from dns.rdtypes.IN.A import A
from dns.rdtypes.IN.AAAA import AAAA
from dns.rrset import RRset

from chik.seeder.dns_server import DNSServer
from chik.seeder.peer_record import PeerRecord, PeerReliability
from chik.util.ints import uint32, uint64

timeout = 0.5


def generate_test_combs() -> list[tuple[bool, str, dns.rdatatype.RdataType]]:
    """
    Generates all the combinations of tests we want to run.
    """
    output = []
    use_tcp = [True, False]
    target_address = ["::1", "127.0.0.1"]
    request_types = [dns.rdatatype.A, dns.rdatatype.AAAA, dns.rdatatype.ANY, dns.rdatatype.NS, dns.rdatatype.SOA]
    # (use_tcp, target-addr, request_type), so udp + tcp, ipv6 +4 for every request type we support.
    for addr in target_address:
        for tcp in use_tcp:
            for req in request_types:
                output.append((tcp, addr, req))
    return output


def get_dns_port(dns_server: DNSServer, use_tcp: bool, target_addr: str) -> int:
    if use_tcp:
        assert dns_server.tcp_server is not None
        tcp_sockets = dns_server.tcp_server.sockets
        wanted_addr = "::" if target_addr == "::1" else "0.0.0.0"
        if tcp_sockets[0].getsockname()[0] == wanted_addr:
            return int(tcp_sockets[0].getsockname()[1])
        return int(tcp_sockets[1].getsockname()[1])
    return dns_server.udp_dns_port


all_test_combinations = generate_test_combs()


@dataclass(frozen=True)
class FakeDnsPacket:
    real_packet: dns.message.Message

    def to_wire(self) -> bytes:
        return self.real_packet.to_wire()[23:]

    def __getattr__(self, item: object) -> None:
        # This is definitely cheating, but it works
        return None


@pytest.fixture(scope="module")
def database_peers() -> dict[str, PeerReliability]:
    """
    We override the values in the class with these dbs, to save time.
    """
    host_to_reliability = {}
    ipv4, ipv6 = get_addresses(2)  # get 1000 addresses
    # add peers to the in memory part of the db.
    for peer in [str(peer) for pair in zip(ipv4, ipv6) for peer in pair]:
        new_peer = PeerRecord(
            peer,
            peer,
            uint32(59678),
            False,
            uint64(0),
            uint32(0),
            uint64(0),
            uint64(int(time.time())),
            uint64(0),
            "undefined",
            uint64(0),
            tls_version="unknown",
        )
        new_peer_reliability = PeerReliability(peer, tries=3, successes=3)  # make sure the peer starts as reliable.
        host_to_reliability[new_peer.peer_id] = new_peer_reliability
    return host_to_reliability


async def make_dns_query(
    use_tcp: bool, target_address: str, port: int, dns_message: dns.message.Message, d_timeout: float = timeout
) -> dns.message.Message:
    """
    Makes a DNS query for the given domain name using the given protocol type.
    """
    if use_tcp:
        return await dns.asyncquery.tcp(q=dns_message, where=target_address, timeout=d_timeout, port=port)
    return await dns.asyncquery.udp(q=dns_message, where=target_address, timeout=d_timeout, port=port)


def get_addresses(num_subnets: int = 10) -> tuple[list[IPv4Address], list[IPv6Address]]:
    ipv4 = []
    ipv6 = []
    # generate 2500 ipv4 and 2500 ipv6 peers, it's just a string so who cares
    for s in range(num_subnets):
        for i in range(1, 251):  # im being lazy as we can only have 255 per subnet
            ipv4.append(IPv4Address(f"192.168.{s}.{i}"))
            ipv6.append(IPv6Address(f"2001:db8::{s}:{i}"))
    return ipv4, ipv6


def assert_standard_results(
    std_query_answer: list[dns.rrset.RRset], request_type: dns.rdatatype.RdataType, num_ns: int
) -> None:
    if request_type == dns.rdatatype.A:
        assert len(std_query_answer) == 1  # only 1 kind of answer
        a_answer = std_query_answer[0]
        assert a_answer.rdtype == dns.rdatatype.A
        assert len(a_answer) == 32  # 32 ipv4 addresses
    elif request_type == dns.rdatatype.AAAA:
        assert len(std_query_answer) == 1  # only 1 kind of answer
        aaaa_answer = std_query_answer[0]
        assert aaaa_answer.rdtype == dns.rdatatype.AAAA
        assert len(aaaa_answer) == 32  # 32 ipv6 addresses
    elif request_type == dns.rdatatype.ANY:
        assert len(std_query_answer) == 4  # 4 kinds of answers
        for answer in std_query_answer:
            if answer.rdtype == dns.rdatatype.A:
                assert len(answer) == 16
            elif answer.rdtype == dns.rdatatype.AAAA:
                assert len(answer) == 16
            elif answer.rdtype == dns.rdatatype.NS:
                assert len(answer) == num_ns
            else:
                assert len(answer) == 1
    elif request_type == dns.rdatatype.NS:
        assert len(std_query_answer) == 1  # only 1 kind of answer
        ns_answer = std_query_answer[0]
        assert ns_answer.rdtype == dns.rdatatype.NS
        assert len(ns_answer) == num_ns  # ns records
    else:
        assert len(std_query_answer) == 1  # soa
        soa_answer = std_query_answer[0]
        assert soa_answer.rdtype == dns.rdatatype.SOA
        assert len(soa_answer) == 1


@pytest.mark.skip(reason="Flaky test with fixes in progress")
@pytest.mark.anyio
@pytest.mark.parametrize("use_tcp, target_address, request_type", all_test_combinations)
async def test_error_conditions(
    seeder_service: DNSServer, use_tcp: bool, target_address: str, request_type: dns.rdatatype.RdataType
) -> None:
    """
    We check having no peers, an invalid packet, an early EOF, and a packet then an EOF halfway through (tcp only).
    We also check for a dns record that does not exist, and a dns record outside the domain.
    """
    port = get_dns_port(seeder_service, use_tcp, target_address)
    domain = seeder_service.domain  # default is: seeder.example.com
    num_ns = len(seeder_service.ns_records)

    # No peers
    no_peers = dns.message.make_query(domain, request_type)
    no_peers_response = await make_dns_query(use_tcp, target_address, port, no_peers)
    assert no_peers_response.rcode() == dns.rcode.NOERROR

    if request_type in {dns.rdatatype.A, dns.rdatatype.AAAA}:
        assert len(no_peers_response.answer) == 0  # no response, as expected
    elif request_type == dns.rdatatype.ANY:  # ns + soa
        assert len(no_peers_response.answer) == 2
        for answer in no_peers_response.answer:
            if answer.rdtype == dns.rdatatype.NS:
                assert len(answer.items) == num_ns
            else:
                assert len(answer.items) == 1
    elif request_type == dns.rdatatype.NS:
        assert len(no_peers_response.answer) == 1  # ns
        assert no_peers_response.answer[0].rdtype == dns.rdatatype.NS
        assert len(no_peers_response.answer[0].items) == num_ns
    else:
        assert len(no_peers_response.answer) == 1  # soa
        assert no_peers_response.answer[0].rdtype == dns.rdatatype.SOA
        assert len(no_peers_response.answer[0].items) == 1
    # Authority Records
    assert len(no_peers_response.authority) == num_ns + 1  # ns + soa
    for record_list in no_peers_response.authority:
        if record_list.rdtype == dns.rdatatype.NS:
            assert len(record_list.items) == num_ns
        else:
            assert len(record_list.items) == 1  # soa

    # Invalid packet (this is kinda a pain)
    invalid_packet = cast(dns.message.Message, FakeDnsPacket(dns.message.make_query(domain, request_type)))
    with pytest.raises(EOFError if use_tcp else dns.exception.Timeout):  # UDP will time out, TCP will EOF
        await make_dns_query(use_tcp, target_address, port, invalid_packet)

    # early EOF packet
    proto, source_addr = (AF_INET6, "::") if target_address == "::1" else (AF_INET, "0.0.0.0")
    if use_tcp:
        backend = dns.asyncbackend.get_default_backend()
        tcp_socket = await backend.make_socket(  # type: ignore[no-untyped-call]
            proto, SOCK_STREAM, 0, (source_addr, 0), (target_address, port), timeout=timeout
        )
        async with tcp_socket as socket:
            await socket.close()
            with pytest.raises(EOFError):
                await dns.asyncquery.receive_tcp(tcp_socket, timeout)

    # Packet then length header then EOF
    if use_tcp:
        backend = dns.asyncbackend.get_default_backend()
        tcp_socket = await backend.make_socket(  # type: ignore[no-untyped-call]
            proto, SOCK_STREAM, 0, (source_addr, 0), (target_address, port), timeout=timeout
        )
        async with tcp_socket as socket:  # send packet then length then eof
            r = await dns.asyncquery.tcp(q=no_peers, where=target_address, timeout=timeout, sock=socket)
            assert r.answer == no_peers_response.answer
            # send 120, as the first 2 bytes / the length of the packet, so that the server expects more.
            await socket.sendall((120).to_bytes(2, byteorder="big"), int(time.time() + timeout))
            await socket.close()
            with pytest.raises(EOFError):
                await dns.asyncquery.receive_tcp(tcp_socket, timeout)

    # Record does not exist
    record_does_not_exist = dns.message.make_query("doesnotexist." + domain, request_type)
    record_does_not_exist_response = await make_dns_query(use_tcp, target_address, port, record_does_not_exist)
    assert record_does_not_exist_response.rcode() == dns.rcode.NXDOMAIN
    assert len(record_does_not_exist_response.answer) == 0
    assert len(record_does_not_exist_response.authority) == num_ns + 1  # ns + soa

    # Record outside domain
    record_outside_domain = dns.message.make_query("chiknetwork.com", request_type)
    record_outside_domain_response = await make_dns_query(use_tcp, target_address, port, record_outside_domain)
    assert record_outside_domain_response.rcode() == dns.rcode.REFUSED
    assert len(record_outside_domain_response.answer) == 0
    assert len(record_outside_domain_response.authority) == 0


@pytest.mark.skip(reason="Flaky test with fixes in progress")
@pytest.mark.anyio
@pytest.mark.parametrize("use_tcp, target_address, request_type", all_test_combinations)
async def test_dns_queries(
    seeder_service: DNSServer, use_tcp: bool, target_address: str, request_type: dns.rdatatype.RdataType
) -> None:
    """
    We add 5000 peers directly, then try every kind of query many times over both the TCP and UDP protocols.
    """
    port = get_dns_port(seeder_service, use_tcp, target_address)
    domain = seeder_service.domain  # default is: seeder.example.com
    num_ns = len(seeder_service.ns_records)

    # add 5000 peers (2500 ipv4, 2500 ipv6)
    seeder_service.reliable_peers_v4, seeder_service.reliable_peers_v6 = get_addresses()

    # now we query for each type of record a lot of times and make sure we get the right number of responses
    for i in range(150):
        query = dns.message.make_query(domain, request_type, use_edns=True)  # we need to generate a new request id.
        std_query_response = await make_dns_query(use_tcp, target_address, port, query)
        assert std_query_response.rcode() == dns.rcode.NOERROR
        assert_standard_results(std_query_response.answer, request_type, num_ns)

        # Assert Authority Records
        assert len(std_query_response.authority) == num_ns + 1  # ns + soa
        for record_list in std_query_response.authority:
            if record_list.rdtype == dns.rdatatype.NS:
                assert len(record_list.items) == num_ns
            else:
                assert len(record_list.items) == 1  # soa
    if not use_tcp:
        # Validate EDNS
        e_query = dns.message.make_query(domain, dns.rdatatype.ANY, use_edns=False)
        with pytest.raises(dns.query.BadResponse):  # response is truncated without EDNS
            await make_dns_query(use_tcp, target_address, port, e_query)


@pytest.mark.anyio
@pytest.mark.parametrize("use_tcp, target_address, request_type", all_test_combinations)
async def test_db_processing(
    seeder_service: DNSServer,
    database_peers: dict[str, PeerReliability],
    use_tcp: bool,
    target_address: str,
    request_type: dns.rdatatype.RdataType,
) -> None:
    """
    We add 1000 peers through the db, then try every kind of query over both the TCP and UDP protocols.
    """
    port = get_dns_port(seeder_service, use_tcp, target_address)
    domain = seeder_service.domain  # default is: seeder.example.com
    num_ns = len(seeder_service.ns_records)
    crawl_store = seeder_service.crawl_store
    assert crawl_store is not None

    # override host_to_reliability with the pre-generated db peers
    crawl_store.host_to_reliability = database_peers

    # Write these new peers to db.
    await crawl_store.load_reliable_peers_to_db()

    # Force a DB reload and ensure it loaded the peers
    await seeder_service.refresh_reliable_peers()
    assert seeder_service.reliable_peers_v4 != []

    # now we check that the db peers are being used.
    query = dns.message.make_query(domain, request_type, use_edns=True)  # we need to generate a new request id.
    std_query_response = await make_dns_query(use_tcp, target_address, port, query)
    assert std_query_response.rcode() == dns.rcode.NOERROR
    assert_standard_results(std_query_response.answer, request_type, num_ns)


@pytest.mark.anyio
@pytest.mark.parametrize(
    "use_tcp, target_address, request_type",
    [
        [False, "127.0.0.1", dns.rdatatype.A],
        [False, "127.0.0.1", dns.rdatatype.AAAA],
        [True, "127.0.0.1", dns.rdatatype.A],
        [True, "127.0.0.1", dns.rdatatype.AAAA],
        [False, "::1", dns.rdatatype.A],
        [False, "::1", dns.rdatatype.AAAA],
        [True, "::1", dns.rdatatype.A],
        [True, "::1", dns.rdatatype.AAAA],
    ],
)
async def test_static_peers(
    seeder_service: DNSServer,
    database_peers: dict[str, PeerReliability],
    use_tcp: bool,
    target_address: str,
    request_type: dns.rdatatype.RdataType,
) -> None:
    port = get_dns_port(seeder_service, False, target_address)
    domain = seeder_service.domain  # default is: seeder.example.com

    seeder_service.config["static_peers"] = ["1.2.3.4", "2001:0DB8::5"]
    await seeder_service.refresh_reliable_peers()

    query = dns.message.make_query(domain, request_type, use_edns=True)  # we need to generate a new request id.
    std_query_response = await make_dns_query(False, target_address, port, query, d_timeout=10)
    assert std_query_response.rcode() == dns.rcode.NOERROR
    assert len(std_query_response.answer) == 1  # only 1 kind of answer
    answer = std_query_response.answer[0]
    assert answer.rdtype == request_type
    assert len(answer) == 1
    expected = "2001:db8::5" if request_type == dns.rdatatype.AAAA else "1.2.3.4"
    assert answer[0].to_text() == expected


def get_mock_resolver() -> AsyncMock:
    # Mock IPv4 response
    mock_rrset_a = RRset(from_text("node.example.com."), IN, A_TYPE)
    mock_rrset_a.add(A(IN, A_TYPE, "1.2.3.4"))  # type: ignore

    # Mock IPv6 response
    mock_rrset_aaaa = RRset(from_text("node.example.com."), IN, AAAA_TYPE)
    mock_rrset_aaaa.add(AAAA(IN, AAAA_TYPE, "2001:db8::5"))  # type: ignore

    # Create a mock Resolver
    mock_resolver = AsyncMock()

    # Adjust mock_resolve to accept all arguments
    async def mock_resolve(
        qname: Union[dns.name.Name, str],
        rdtype: Union[dns.rdatatype.RdataType, str] = dns.rdatatype.A,
        lifetime: Optional[float] = None,
    ) -> RRset:
        if rdtype == "A":
            return mock_rrset_a
        elif rdtype == "AAAA":
            return mock_rrset_aaaa
        raise Exception(f"Unexpected query type: {rdtype}")

    mock_resolver.resolve.side_effect = mock_resolve
    return mock_resolver


@pytest.mark.anyio
@pytest.mark.parametrize(
    "use_tcp, target_address, request_type",
    [
        [False, "127.0.0.1", dns.rdatatype.A],
        [False, "127.0.0.1", dns.rdatatype.AAAA],
        [True, "127.0.0.1", dns.rdatatype.A],
        [True, "127.0.0.1", dns.rdatatype.AAAA],
        [False, "::1", dns.rdatatype.A],
        [False, "::1", dns.rdatatype.AAAA],
        [True, "::1", dns.rdatatype.A],
        [True, "::1", dns.rdatatype.AAAA],
    ],
)
async def test_static_peers_with_hostnames(
    seeder_service: DNSServer,
    database_peers: dict[str, PeerReliability],
    use_tcp: bool,
    target_address: str,
    request_type: dns.rdatatype.RdataType,
) -> None:
    port = get_dns_port(seeder_service, False, target_address)
    domain = seeder_service.domain  # default is: seeder.example.com

    seeder_service.resolver = get_mock_resolver()
    seeder_service.config["static_peers"] = ["node.example.com"]
    await seeder_service.refresh_reliable_peers()

    query = dns.message.make_query(domain, request_type, use_edns=True)  # we need to generate a new request id.
    std_query_response = await make_dns_query(False, target_address, port, query, d_timeout=10)
    assert std_query_response.rcode() == dns.rcode.NOERROR
    assert len(std_query_response.answer) == 1  # only 1 kind of answer
    answer = std_query_response.answer[0]
    assert answer.rdtype == request_type
    assert len(answer) == 1
    expected = "2001:db8::5" if request_type == dns.rdatatype.AAAA else "1.2.3.4"
    assert answer[0].to_text() == expected



from __future__ import annotations

from dataclasses import dataclass

from chik.util.ints import uint8, uint32
from chik.util.streamable import Streamable, streamable


@streamable
@dataclass(frozen=True)
class WalletInfo(Streamable):
    """
    This object represents the wallet data as it is stored in DB.
    ID: Main wallet (Standard) is stored at index 1, every wallet created after done has auto incremented id.
    Name: can be a user provided or default generated name. (can be modified)
    Type: is specified during wallet creation and should never be changed.
    Data: this filed is intended to be used for storing any wallet specific information required for it.
    (RL wallet stores origin_id, admin/user pubkey, rate limit, etc.)
    This data should be json encoded string.
    """

    id: uint32
    name: str
    type: uint8  # WalletType(type)
    data: str


@streamable
@dataclass(frozen=True)
class WalletInfoBackup(Streamable):
    """
    Used for transforming list of WalletInfo objects into bytes.
    """

    wallet_list: list[WalletInfo]



from __future__ import annotations

import asyncio
import contextlib
import json
import logging
import sys
import traceback
from collections.abc import AsyncIterator, Awaitable
from dataclasses import dataclass
from pathlib import Path
from ssl import SSLContext
from types import MethodType
from typing import Any, Callable, ClassVar, Generic, Optional, TypeVar

from aiohttp import (
    ClientConnectorError,
    ClientSession,
    ClientWebSocketResponse,
    WSMsgType,
    web,
)
from typing_extensions import Protocol, final

from chik import __version__
from chik.rpc.util import wrap_http_handler
from chik.server.outbound_message import NodeType
from chik.server.server import (
    ChikServer,
    ssl_context_for_client,
    ssl_context_for_server,
)
from chik.server.ws_connection import WSChikConnection
from chik.types.peer_info import PeerInfo
from chik.util.byte_types import hexstr_to_bytes
from chik.util.chik_logging import default_log_level, set_log_level
from chik.util.config import str2bool
from chik.util.ints import uint16
from chik.util.json_util import dict_to_json_str
from chik.util.network import WebServer, resolve
from chik.util.task_referencer import create_referenced_task
from chik.util.ws_message import (
    WsRpcMessage,
    create_payload,
    create_payload_dict,
    format_response,
    pong,
)

log = logging.getLogger(__name__)
max_message_size = 50 * 1024 * 1024  # 50MB


EndpointResult = dict[str, Any]
Endpoint = Callable[[dict[str, object]], Awaitable[EndpointResult]]
_T_RpcApiProtocol = TypeVar("_T_RpcApiProtocol", bound="RpcApiProtocol")


class StateChangedProtocol(Protocol):
    def __call__(self, change: str, change_data: Optional[dict[str, Any]]) -> None: ...


class RpcServiceProtocol(Protocol):
    _shut_down: bool
    """Indicates a request to shut down the service.

    This is generally set internally by the class itself and not used externally.
    Consider replacing with asyncio cancellation.
    """

    @property
    def server(self) -> ChikServer:
        """The server object that handles the common server behavior for the RPC."""
        # a property so as to be read only which allows ChikServer to satisfy
        # Optional[ChikServer]
        ...

    def get_connections(self, request_node_type: Optional[NodeType]) -> list[dict[str, Any]]:
        """Report the active connections for the service.

        A default implementation is available and can be called as
        chik.rpc.rpc_server.default_get_connections()
        """
        ...

    async def on_connect(self, peer: WSChikConnection) -> None:
        """Called when a new connection is established to the server."""
        ...

    def _set_state_changed_callback(self, callback: StateChangedProtocol) -> None:
        """Register the callable that will process state change events."""
        ...

    @contextlib.asynccontextmanager
    async def manage(self) -> AsyncIterator[None]:
        yield  # pragma: no cover


class RpcApiProtocol(Protocol):
    service_name: str
    """The name of the service.

    All lower case with underscores as needed.
    """

    def __init__(self, node: RpcServiceProtocol) -> None: ...

    @property
    def service(self) -> RpcServiceProtocol:
        """The service object that provides the specific behavior for the API."""
        # using a read-only property per https://github.com/python/mypy/issues/12990
        ...

    def get_routes(self) -> dict[str, Endpoint]:
        """Return the mapping of endpoints to handler callables."""
        ...

    async def _state_changed(self, change: str, change_data: Optional[dict[str, Any]]) -> list[WsRpcMessage]:
        """Notify the state change system of a changed state."""
        ...


def default_get_connections(server: ChikServer, request_node_type: Optional[NodeType]) -> list[dict[str, Any]]:
    connections = server.get_connections(request_node_type)
    con_info = [
        {
            "type": con.connection_type,
            "local_port": con.local_port,
            "peer_host": con.peer_info.host,
            "peer_port": con.peer_info.port,
            "peer_server_port": con.peer_server_port,
            "node_id": con.peer_node_id,
            "creation_time": con.creation_time,
            "bytes_read": con.bytes_read,
            "bytes_written": con.bytes_written,
            "last_message_time": con.last_message_time,
        }
        for con in connections
    ]
    return con_info


@final
@dataclass
class RpcServer(Generic[_T_RpcApiProtocol]):
    """
    Implementation of RPC server.
    """

    rpc_api: _T_RpcApiProtocol
    stop_cb: Callable[[], None]
    service_name: str
    ssl_context: SSLContext
    ssl_client_context: SSLContext
    net_config: dict[str, Any]
    service_config: dict[str, Any]
    webserver: Optional[WebServer] = None
    daemon_heartbeat: int = 300
    daemon_connection_task: Optional[asyncio.Task[None]] = None
    shut_down: bool = False
    websocket: Optional[ClientWebSocketResponse] = None
    client_session: Optional[ClientSession] = None
    prefer_ipv6: bool = False

    @classmethod
    def create(
        cls,
        rpc_api: _T_RpcApiProtocol,
        service_name: str,
        stop_cb: Callable[[], None],
        root_path: Path,
        net_config: dict[str, Any],
        service_config: dict[str, Any],
        prefer_ipv6: bool,
    ) -> RpcServer[_T_RpcApiProtocol]:
        crt_path = root_path / net_config["daemon_ssl"]["private_crt"]
        key_path = root_path / net_config["daemon_ssl"]["private_key"]
        ca_cert_path = root_path / net_config["private_ssl_ca"]["crt"]
        ca_key_path = root_path / net_config["private_ssl_ca"]["key"]
        daemon_heartbeat = net_config.get("daemon_heartbeat", 300)
        ssl_context = ssl_context_for_server(ca_cert_path, ca_key_path, crt_path, key_path, log=log)
        ssl_client_context = ssl_context_for_client(ca_cert_path, ca_key_path, crt_path, key_path, log=log)
        return cls(
            rpc_api,
            stop_cb,
            service_name,
            ssl_context,
            ssl_client_context,
            net_config,
            service_config=service_config,
            daemon_heartbeat=daemon_heartbeat,
            prefer_ipv6=prefer_ipv6,
        )

    async def start(self, self_hostname: str, rpc_port: uint16, max_request_body_size: int) -> None:
        if self.webserver is not None:
            raise RuntimeError("RpcServer already started")
        self.webserver = await WebServer.create(
            hostname=self_hostname,
            port=rpc_port,
            max_request_body_size=max_request_body_size,
            routes=[web.post(route, wrap_http_handler(func)) for (route, func) in self._get_routes().items()],
            ssl_context=self.ssl_context,
            prefer_ipv6=self.prefer_ipv6,
        )

    def close(self) -> None:
        self.shut_down = True
        if self.webserver is not None:
            self.webserver.close()

    async def await_closed(self) -> None:
        if self.websocket is not None:
            await self.websocket.close()
        if self.client_session is not None:
            await self.client_session.close()
        if self.webserver is not None:
            await self.webserver.await_closed()
        if self.daemon_connection_task is not None:
            await self.daemon_connection_task
            self.daemon_connection_task = None

    async def _state_changed(self, change: str, change_data: Optional[dict[str, Any]]) -> None:
        if self.websocket is None or self.websocket.closed:
            return None
        payloads: list[WsRpcMessage] = await self.rpc_api._state_changed(change, change_data)

        if change in {"add_connection", "close_connection", "peer_changed_peak"}:
            data = await self.get_connections({})
            if data is not None:
                payload = create_payload_dict(
                    "get_connections",
                    data,
                    self.service_name,
                    "wallet_ui",
                )
                payloads.append(payload)
        for payload in payloads:
            if "success" not in payload["data"]:
                payload["data"]["success"] = True
            if self.websocket is None or self.websocket.closed:
                return None
            try:
                await self.websocket.send_str(dict_to_json_str(payload))
            except Exception:
                tb = traceback.format_exc()
                log.warning(f"Sending data failed. Exception {tb}.")

    def state_changed(self, change: str, change_data: Optional[dict[str, Any]] = None) -> None:
        if self.websocket is None or self.websocket.closed:
            return None
        create_referenced_task(self._state_changed(change, change_data), known_unreferenced=True)

    @property
    def listen_port(self) -> uint16:
        if self.webserver is None:
            raise RuntimeError("RpcServer is not started")
        return self.webserver.listen_port

    def _get_routes(self) -> dict[str, Endpoint]:
        return {
            **self.rpc_api.get_routes(),
            **{path: MethodType(handler, self) for path, handler in self._routes.items()},
        }

    async def get_routes(self, request: dict[str, Any]) -> EndpointResult:
        return {
            "success": True,
            "routes": list(self._get_routes().keys()),
        }

    async def get_network_info(self, _: dict[str, Any]) -> EndpointResult:
        network_name = self.net_config["selected_network"]
        address_prefix = self.net_config["network_overrides"]["config"][network_name]["address_prefix"]
        genesis_challenge = self.net_config["network_overrides"]["constants"][network_name]["GENESIS_CHALLENGE"]
        return {
            "network_name": network_name,
            "network_prefix": address_prefix,
            "genesis_challenge": genesis_challenge,
        }

    async def get_connections(self, request: dict[str, Any]) -> EndpointResult:
        request_node_type: Optional[NodeType] = None
        if "node_type" in request:
            request_node_type = NodeType(request["node_type"])
        if self.rpc_api.service.server is None:
            raise ValueError("Global connections is not set")
        con_info: list[dict[str, Any]]
        con_info = self.rpc_api.service.get_connections(request_node_type=request_node_type)
        return {"connections": con_info}

    async def open_connection(self, request: dict[str, Any]) -> EndpointResult:
        host = request["host"]
        port = request["port"]
        target_node: PeerInfo = PeerInfo(await resolve(host, prefer_ipv6=self.prefer_ipv6), uint16(int(port)))
        on_connect = None
        if hasattr(self.rpc_api.service, "on_connect"):
            on_connect = self.rpc_api.service.on_connect
        if not await self.rpc_api.service.server.start_client(target_node, on_connect):
            return {"success": False, "error": f"could not connect to {target_node}"}
        return {"success": True}

    async def close_connection(self, request: dict[str, Any]) -> EndpointResult:
        node_id = hexstr_to_bytes(request["node_id"])
        if self.rpc_api.service.server is None:
            raise web.HTTPInternalServerError()
        connections_to_close = [c for c in self.rpc_api.service.server.get_connections() if c.peer_node_id == node_id]
        if len(connections_to_close) == 0:
            raise ValueError(f"Connection with node_id {node_id.hex()} does not exist")
        for connection in connections_to_close:
            await connection.close()
        return {}

    async def stop_node(self, request: dict[str, Any]) -> EndpointResult:
        """
        Shuts down the node.
        """
        if self.stop_cb is not None:
            self.stop_cb()
        return {}

    async def healthz(self, request: dict[str, Any]) -> EndpointResult:
        return {
            "success": True,
        }

    async def get_version(self, request: dict[str, Any]) -> EndpointResult:
        return {
            "version": __version__,
        }

    async def get_log_level(self, request: dict[str, Any]) -> EndpointResult:
        logger = logging.getLogger()
        level_number = logger.level
        level_name = logging.getLevelName(level_number)

        if sys.version_info >= (3, 11):
            map = logging.getLevelNamesMapping()
        else:
            map = logging._nameToLevel

        return {
            "success": True,
            "level": level_name,
            "available_levels": list(map),
        }

    async def reset_log_level(self, request: dict[str, Any]) -> EndpointResult:
        level_name = self.service_config.get("log_level", default_log_level)

        return await self.set_log_level(request={"level": level_name})

    async def set_log_level(self, request: dict[str, Any]) -> EndpointResult:
        error_strings = set_log_level(log_level=request["level"], service_name=self.service_name)
        status = await self.get_log_level(request={})

        status["success"] &= len(error_strings) == 0

        return {
            **status,
            "errors": error_strings,
        }

    async def ws_api(self, message: WsRpcMessage) -> Optional[dict[str, object]]:
        """
        This function gets called when new message is received via websocket.
        """

        command = message["command"]
        if message["ack"]:
            return None

        data: dict[str, object] = {}
        if "data" in message:
            data = message["data"]
        if command == "ping":
            return pong()

        f_internal: Optional[Endpoint] = getattr(self, command, None)
        if f_internal is not None:
            return await f_internal(data)
        f_rpc_api: Optional[Endpoint] = getattr(self.rpc_api, command, None)
        if f_rpc_api is not None:
            return await f_rpc_api(data)

        raise ValueError(f"unknown_command {command}")

    async def safe_handle(self, websocket: ClientWebSocketResponse, payload: str) -> None:
        message = None
        try:
            message = json.loads(payload)
            log.debug(f"Rpc call <- {message['command']}")
            response = await self.ws_api(message)

            # Only respond if we return something from api call
            if response is not None:
                log.debug(f"Rpc response -> {message['command']}")
                # Set success to true automatically (unless it's already set)
                if "success" not in response:
                    response["success"] = True
                await websocket.send_str(format_response(message, response))

        except Exception as e:
            tb = traceback.format_exc()
            log.warning(f"Error while handling message: {tb}")
            if message is not None:
                error = e.args[0] if e.args else e
                res = {"success": False, "error": f"{error}"}
                await websocket.send_str(format_response(message, res))

    async def connection(self, ws: ClientWebSocketResponse) -> None:
        data = {"service": self.service_name}
        payload = create_payload("register_service", data, self.service_name, "daemon")
        await ws.send_str(payload)

        while True:
            # ClientWebSocketReponse::receive() internally handles PING, PONG, and CLOSE messages
            msg = await ws.receive()
            if msg.type == WSMsgType.TEXT:
                message = msg.data.strip()
                # log.info(f"received message: {message}")
                await self.safe_handle(ws, message)
            elif msg.type == WSMsgType.BINARY:
                log.debug("Received binary data")
            else:
                if msg.type == WSMsgType.ERROR:
                    log.error("Error during receive %s", ws.exception())
                elif msg.type == WSMsgType.CLOSED:
                    pass

                break

    def connect_to_daemon(self, self_hostname: str, daemon_port: uint16) -> None:
        if self.daemon_connection_task is not None:
            raise RuntimeError("Already connected to the daemon")

        async def inner() -> None:
            while not self.shut_down:
                try:
                    self.client_session = ClientSession()
                    self.websocket = await self.client_session.ws_connect(
                        f"wss://{self_hostname}:{daemon_port}",
                        autoclose=True,
                        autoping=True,
                        heartbeat=self.daemon_heartbeat,
                        ssl=self.ssl_client_context,
                        max_msg_size=max_message_size,
                    )
                    await self.connection(self.websocket)
                except ClientConnectorError:
                    log.warning(f"Cannot connect to daemon at ws://{self_hostname}:{daemon_port}")
                except Exception as e:
                    tb = traceback.format_exc()
                    log.warning(f"Exception: {tb} {type(e)}")
                if self.websocket is not None:
                    await self.websocket.close()
                if self.client_session is not None:
                    await self.client_session.close()
                self.websocket = None
                self.client_session = None
                await asyncio.sleep(2)

        self.daemon_connection_task = create_referenced_task(inner())

    _routes: ClassVar[dict[str, Callable[..., Awaitable[object]]]] = {
        "/get_network_info": get_network_info,
        "/get_connections": get_connections,
        "/open_connection": open_connection,
        "/close_connection": close_connection,
        "/stop_node": stop_node,
        "/get_routes": get_routes,
        "/get_version": get_version,
        "/healthz": healthz,
        "/get_log_level": get_log_level,
        "/set_log_level": set_log_level,
        "/reset_log_level": reset_log_level,
    }


async def start_rpc_server(
    rpc_api: _T_RpcApiProtocol,
    self_hostname: str,
    daemon_port: uint16,
    rpc_port: uint16,
    stop_cb: Callable[[], None],
    root_path: Path,
    net_config: dict[str, object],
    service_config: dict[str, object],
    connect_to_daemon: bool = True,
    max_request_body_size: Optional[int] = None,
) -> RpcServer[_T_RpcApiProtocol]:
    """
    Starts an HTTP server with the following RPC methods, to be used by local clients to
    query the node.
    """
    try:
        if max_request_body_size is None:
            max_request_body_size = 1024**2

        prefer_ipv6 = str2bool(str(net_config.get("prefer_ipv6", False)))

        rpc_server = RpcServer.create(
            rpc_api,
            rpc_api.service_name,
            stop_cb,
            root_path,
            net_config,
            service_config=service_config,
            prefer_ipv6=prefer_ipv6,
        )
        rpc_server.rpc_api.service._set_state_changed_callback(rpc_server.state_changed)
        await rpc_server.start(self_hostname, rpc_port, max_request_body_size)

        if connect_to_daemon:
            rpc_server.connect_to_daemon(self_hostname, daemon_port)

        return rpc_server
    except Exception:
        tb = traceback.format_exc()
        log.error(f"Starting RPC server failed. Exception {tb}")
        raise



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Optional, Union

from klvm_tools.binutils import disassemble

from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.wallet.puzzle_drivers import PuzzleInfo, Solver
from chik.wallet.puzzles.load_klvm import load_klvm_maybe_recompile
from chik.wallet.uncurried_puzzle import UncurriedPuzzle, uncurry_puzzle

OWNERSHIP_LAYER_MOD = load_klvm_maybe_recompile(
    "nft_ownership_layer.clsp", package_or_requirement="chik.wallet.nft_wallet.puzzles"
)


def match_ownership_layer_puzzle(puzzle: UncurriedPuzzle) -> tuple[bool, list[Program]]:
    if puzzle.mod == OWNERSHIP_LAYER_MOD:
        return True, list(puzzle.args.as_iter())
    return False, []


def puzzle_for_ownership_layer(
    current_owner: Union[Program, bytes], transfer_program: Program, inner_puzzle: Program
) -> Program:
    return OWNERSHIP_LAYER_MOD.curry(OWNERSHIP_LAYER_MOD.get_tree_hash(), current_owner, transfer_program, inner_puzzle)


def solution_for_ownership_layer(inner_solution: Program) -> Program:
    return Program.to([inner_solution])


@dataclass(frozen=True)
class OwnershipOuterPuzzle:
    _match: Callable[[UncurriedPuzzle], Optional[PuzzleInfo]]
    _construct: Callable[[PuzzleInfo, Program], Program]
    _solve: Callable[[PuzzleInfo, Solver, Program, Program], Program]
    _get_inner_puzzle: Callable[[PuzzleInfo, UncurriedPuzzle], Optional[Program]]
    _get_inner_solution: Callable[[PuzzleInfo, Program], Optional[Program]]

    def match(self, puzzle: UncurriedPuzzle) -> Optional[PuzzleInfo]:
        matched, curried_args = match_ownership_layer_puzzle(puzzle)
        if matched:
            _, current_owner, transfer_program, inner_puzzle = curried_args
            owner_bytes: bytes = current_owner.as_python()
            tp_match: Optional[PuzzleInfo] = self._match(uncurry_puzzle(transfer_program))
            constructor_dict = {
                "type": "ownership",
                "owner": "()" if owner_bytes == b"" else "0x" + owner_bytes.hex(),
                "transfer_program": (disassemble(transfer_program) if tp_match is None else tp_match.info),
            }
            next_constructor = self._match(uncurry_puzzle(inner_puzzle))
            if next_constructor is not None:
                constructor_dict["also"] = next_constructor.info
            return PuzzleInfo(constructor_dict)
        else:
            return None

    def asset_id(self, constructor: PuzzleInfo) -> Optional[bytes32]:
        return None

    def construct(self, constructor: PuzzleInfo, inner_puzzle: Program) -> Program:
        also = constructor.also()
        if also is not None:
            inner_puzzle = self._construct(also, inner_puzzle)
        transfer_program_info: Union[PuzzleInfo, Program] = constructor["transfer_program"]
        if isinstance(transfer_program_info, Program):
            transfer_program: Program = transfer_program_info
        else:
            transfer_program = self._construct(transfer_program_info, inner_puzzle)
        return puzzle_for_ownership_layer(constructor["owner"], transfer_program, inner_puzzle)

    def get_inner_puzzle(self, constructor: PuzzleInfo, puzzle_reveal: UncurriedPuzzle) -> Optional[Program]:
        matched, curried_args = match_ownership_layer_puzzle(puzzle_reveal)
        if matched:
            _, _, _, inner_puzzle = curried_args
            also = constructor.also()
            if also is not None:
                deep_inner_puzzle: Optional[Program] = self._get_inner_puzzle(also, uncurry_puzzle(inner_puzzle))
                return deep_inner_puzzle
            else:
                return inner_puzzle
        else:
            raise ValueError("This driver is not for the specified puzzle reveal")

    def get_inner_solution(self, constructor: PuzzleInfo, solution: Program) -> Optional[Program]:
        my_inner_solution: Program = solution.first()
        also = constructor.also()
        if also:
            deep_inner_solution: Optional[Program] = self._get_inner_solution(also, my_inner_solution)
            return deep_inner_solution
        else:
            return my_inner_solution

    def solve(self, constructor: PuzzleInfo, solver: Solver, inner_puzzle: Program, inner_solution: Program) -> Program:
        also = constructor.also()
        if also is not None:
            inner_solution = self._solve(also, solver, inner_puzzle, inner_solution)
        return solution_for_ownership_layer(inner_solution)



# Package: utils

from __future__ import annotations

import sqlite3

import aiosqlite


async def lookup_db_version(db: aiosqlite.Connection) -> int:
    try:
        cursor = await db.execute("SELECT * from database_version")
        row = await cursor.fetchone()
        if row is not None and row[0] == 2:
            return 2
        else:
            return 1
    except aiosqlite.OperationalError:
        # expects OperationalError('no such table: database_version')
        return 1


async def set_db_version_async(db: aiosqlite.Connection, version: int) -> None:
    await db.execute("CREATE TABLE database_version(version int)")
    await db.execute("INSERT INTO database_version VALUES (?)", (version,))
    await db.commit()


def set_db_version(db: sqlite3.Connection, version: int) -> None:
    db.execute("CREATE TABLE database_version(version int)")
    db.execute("INSERT INTO database_version VALUES (?)", (version,))
    db.commit()



from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum

from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.ints import uint16, uint64
from chik.util.streamable import Streamable, streamable
from chik.wallet.wallet_puzzle_store import WalletPuzzleStore


@streamable
@dataclass(frozen=True)
class ClawbackMetadata(Streamable):
    time_lock: uint64
    sender_puzzle_hash: bytes32
    recipient_puzzle_hash: bytes32

    async def is_recipient(self, puzzle_store: WalletPuzzleStore) -> bool:
        if await puzzle_store.puzzle_hash_exists(self.sender_puzzle_hash):
            return False
        elif await puzzle_store.puzzle_hash_exists(self.recipient_puzzle_hash):
            return True
        else:
            raise ValueError("Both sender and recipient puzzle hashes not found in puzzle store")


class ClawbackVersion(IntEnum):
    V1 = uint16(1)


@streamable
@dataclass(frozen=True)
class AutoClaimSettings(Streamable):
    enabled: bool = False
    tx_fee: uint64 = uint64(0)
    min_amount: uint64 = uint64(0)
    batch_size: uint16 = uint16(50)
